# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Helper classes for finding prerequisite input datasets during
QuantumGraph generation.
"""

from __future__ import annotations

__all__ = (
    "PrerequisiteBounds",
    "PrerequisiteFinder",
    "PrerequisiteInfo",
    "SkyPixBoundsBuilder",
    "TimespanBuilder",
)

import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import cast

from lsst.daf.butler import (
    Butler,
    DataCoordinate,
    DatasetRef,
    DatasetType,
    DimensionElement,
    Registry,
    SkyPixDimension,
    Timespan,
)
from lsst.daf.butler.registry import MissingDatasetTypeError
from lsst.sphgeom import RangeSet, Region

from .pipeline_graph import DatasetTypeNode, PipelineGraph, ReadEdge, TaskNode


@dataclasses.dataclass
class PrerequisiteInfo:
    """A QuantumGraph-generation helper class that manages the searches for all
    prerequisite input connections for a task.

    Parameters
    ----------
    task_node : `TaskNode`
        The relevant node.
    pipeline_graph : `PipelineGraph`
        The pipeline graph.
    """

    bounds: PrerequisiteBounds
    """Another helper object that manages the spatial/temporal bounds of the
    task's quanta.
    """

    finders: dict[str, PrerequisiteFinder]
    """Mapping of helper objects responsible for a single prerequisite input
    connection.

    Keys are connection names.  Elements of this dictionary should be removed
    by implementations of `QuantumGraphBuilder.process_subgraph` to take
    responsibility for finding them away from the the `QuantumGraphBuilder`
    base class.
    """

    def __init__(self, task_node: TaskNode, pipeline_graph: PipelineGraph):
        self.bounds = PrerequisiteBounds(task_node)
        self.finders = {
            edge.connection_name: PrerequisiteFinder(edge, self.bounds, pipeline_graph)
            for edge in task_node.prerequisite_inputs.values()
        }

    def update_bounds(self) -> None:
        """Inspect the current state of `finders` and update `bounds` to
        reflect the needs of only the finders that remain.
        """
        self.bounds.all_dataset_skypix.clear()
        self.bounds.any_dataset_has_timespan = False
        for finder in self.finders.values():
            self.bounds.all_dataset_skypix.update(finder.dataset_skypix)
            self.bounds.any_dataset_has_timespan = (
                self.bounds.any_dataset_has_timespan or finder.dataset_has_timespan
            )


class PrerequisiteFinder:
    """A QuantumGraph-generation helper class that manages the searches for a
    prerequisite input connection.

    Parameters
    ----------
    edge : `pipeline_graph.ReadEdge`
        A `~pipeline_graph.PipelineGraph` edge that represents a single
        prerequisite input connection.
    bounds : `PrerequisiteBounds`
        Another helper object that manages the spatial/temporal bounds of the
        task's quanta, shared by all prerequisite inputs for that task.
    pipeline_graph : `pipeline_graph.PipelineGraph`
        Graph representation of the pipeline.

    Notes
    -----
    `PrerequisiteFinder` instances are usually constructed by a
    `PrerequisiteInfo` instance, which is in turn constructed by and attached
    to the base `QuantumGraphBuilder` when a new builder is constructed. During
    the `QuantumGraphBuilder.process_subgraph` hook implemented by a builder
    subclass, prerequisite inputs may be found in other ways (e.g. via bulk
    queries), as long as the results are consistent with the finder's
    attributes, and this is indicated to the base `QuantumGraphBuilder` by
    removing those finder instances after those prerequisites have been found
    and added to a `QuantumGraphSkeleton`.  Finder instances that remain in the
    builder are used by calling `PrerequisiteFinder.find` on each quantum
    later in `QuantumGraphBuilder.build`.
    """

    def __init__(
        self,
        edge: ReadEdge,
        bounds: PrerequisiteBounds,
        pipeline_graph: PipelineGraph,
    ):
        self.edge = edge
        self._bounds = bounds
        self.dataset_type_node = pipeline_graph.dataset_types[edge.parent_dataset_type_name]
        self.lookup_function = self.task_node.get_lookup_function(edge.connection_name)
        self.dataset_skypix = {}
        self.dataset_other_spatial = {}
        self.dataset_has_timespan = False
        self.constraint_dimensions = self.task_node.dimensions
        if self.lookup_function is None:
            for family in self.dataset_type_node.dimensions.spatial - self.task_node.dimensions.spatial:
                best_spatial_element = family.choose(self.dataset_type_node.dimensions)
                if isinstance(best_spatial_element, SkyPixDimension):
                    self.dataset_skypix[best_spatial_element.name] = best_spatial_element
                else:
                    self.dataset_other_spatial[best_spatial_element.name] = cast(
                        DimensionElement, best_spatial_element
                    )
            self.dataset_has_timespan = bool(
                # If the task dimensions has a temporal family that isn't in
                # the dataset type (i.e. "observation_timespans", like visit
                # or exposure)...
                self.task_node.dimensions.temporal - self.dataset_type_node.dimensions.temporal
            ) and (
                # ...and the dataset type has a temporal family that isn't in
                # the task dimensions, or is a calibration, the prerequisite
                # search needs a temporal join.  Note that the default
                # dimension universe only has one temporal dimension family, so
                # in practice this just means "calibration lookups when visit
                # or exposure is in the task dimensions".
                self.dataset_type_node.is_calibration
                or bool(self.dataset_type_node.dimensions.temporal - self.task_node.dimensions.temporal)
            )
            new_constraint_dimensions = set()
            universe = self.task_node.dimensions.universe
            for dimension_name in self.task_node.dimensions.names:
                if dimension_name in self.dataset_type_node.dimensions.names:
                    new_constraint_dimensions.add(dimension_name)
                else:
                    dimension = universe[dimension_name]
                    if not (dimension.spatial or dimension.temporal):
                        new_constraint_dimensions.add(dimension_name)
            self.constraint_dimensions = universe.conform(new_constraint_dimensions)

    edge: ReadEdge
    """The `~pipeline_graph.PipelineGraph` edge that represents the
    prerequisite input connection.
    """

    dataset_type_node: DatasetTypeNode
    """The `~pipeline_graph.PipelineGraph` node that represents the dataset
    type of this connection.

    This always uses the registry storage class and is never a component
    dataset type.
    """

    lookup_function: (
        Callable[[DatasetType, Registry, DataCoordinate, Sequence[str]], Iterable[DatasetRef]] | None
    )
    """A task-provided callback for finding these datasets.

    If this is not `None`, it must be used to ensure correct behavior.
    """

    dataset_skypix: dict[str, SkyPixDimension]
    """Dimensions representing a pixelization of the sky used by the dataset
    type for this connection that are also not part of the task's dimensions.

    Keys are dimension names.  It is at least extremely rare for this
    dictionary to have more than one element.
    """

    dataset_other_spatial: dict[str, DimensionElement]
    """Spatial dimensions other than sky pixelizations used by the dataset type
    for this connection that are also not part of the task's dimensions.
    """

    dataset_has_timespan: bool
    """Whether the dataset has a timespan that should be used in the lookup,
    either because it is a calibration dataset or because it has temporal
    dimensions that are not part of the tasks's dimensions.
    """

    @property
    def task_node(self) -> TaskNode:
        """The `~pipeline_graph.PipelineGraph` node that represents the task
        for this connection.
        """
        return self._bounds.task_node

    def find(
        self,
        butler: Butler,
        input_collections: Sequence[str],
        data_id: DataCoordinate,
        skypix_bounds: Mapping[str, RangeSet],
        timespan: Timespan | None,
    ) -> list[DatasetRef]:
        """Find prerequisite input datasets for a single quantum.

        Parameters
        ----------
        butler : `lsst.daf.butler.Butler`
            Butler client to use for queries.
        input_collections : `~collections.abc.Sequence` [ `str` ]
            Sequence of collections to search, in order.
        data_id : `lsst.daf.butler.DataCoordinate`
            Data ID for the quantum.
        skypix_bounds : `~collections.abc.Mapping` \
              [ `str`, `lsst.sphgeom.RangeSet` ]
            The spatial bounds of this quantum in various skypix dimensions.
            Keys are skypix dimension names (a superset of those in
            `dataset_skypix`) and values are sets of integer pixel ID ranges.
        timespan : `lsst.daf.butler.Timespan` or `None`
            The temporal bounds of this quantum.  Guaranteed to not be `None`
            if `dataset_has_timespan` is `True`.

        Returns
        -------
        refs : `list` [ `lsst.daf.butler.DatasetRef` ]
            Dataset references.  These use
            ``self.dataset_type_node.dataset_type``, which may differ from the
            connection's dataset type in storage class or [lack of] component.

        Raises
        ------
        NotImplementedError
            Raised for certain relationships between task and dataset type
            dimensions that are possible to define but not believed to be
            useful in practice.  These errors occur late rather than early in
            order to allow a `QuantumGraphBuilder` subclass to handle them
            first, in case an unusual task's needs must be met by a custom
            builder class anyway.
        """
        if self.lookup_function:
            # If there is a lookup function, just use it; nothing else matters.
            return [
                self.dataset_type_node.generalize_ref(ref)
                for ref in self.lookup_function(
                    self.edge.adapt_dataset_type(self.dataset_type_node.dataset_type),
                    butler.registry,
                    data_id,
                    input_collections,
                )
                if ref is not None
            ]
        if self.dataset_type_node.dimensions <= self.constraint_dimensions:
            # If this is a calibration dataset and the dataset doesn't have
            # any dimensions that aren't constrained by the quantum data
            # ID, we know there'll only be one result, and that means we
            # can call Butler.find_dataset, which takes a timespan. Note
            # that the AllDimensionsQuantumGraphBuilder subclass will
            # intercept this case in order to optimize it when:
            #
            #  - PipelineTaskConnections.getTemporalBoundsConnections is
            #    empty;
            #
            #  - the quantum data IDs have temporal dimensions;
            #
            # and when that happens PrerequisiteFinder.find never gets
            # called.
            try:
                ref = butler.find_dataset(
                    self.dataset_type_node.dataset_type,
                    data_id.subset(self.constraint_dimensions),
                    collections=input_collections,
                    timespan=timespan,
                )
            except MissingDatasetTypeError:
                ref = None
            return [ref] if ref is not None else []
        elif self.dataset_has_timespan:
            extra_dimensions = self.dataset_type_node.dimensions.names - self.constraint_dimensions.names
            raise NotImplementedError(
                f"No support for calibration lookup {self.task_node.label}.{self.edge.connection_name} "
                f"with dimension(s) {extra_dimensions} not fully constrained by the task. "
                "Please create a feature-request ticket and use a lookup function in the meantime."
            )
        if self.dataset_skypix:
            if not self.dataset_has_timespan and not self.dataset_other_spatial:
                # If the dataset has skypix dimensions but is not otherwise
                # spatial or temporal (this describes reference catalogs and
                # things like them), we can stuff the skypix IDs we want into
                # the query via bind parameters and call queryDatasets.  Once
                # again AllDimensionsQuantumGraphBuilder will often intercept
                # this case in order to optimize it, when:
                #
                #  - PipelineTaskConnections.getSpatialBoundsConnections is
                #    empty;
                #
                #  - the quantum data IDs have spatial dimensions;
                #
                # and when that happens PrerequisiteFinder.find never gets
                # called.
                where_terms: list[str] = []
                bind: dict[str, list[int]] = {}
                for name in self.dataset_skypix:
                    where_terms.append(f"{name} IN (:{name}_pixels)")
                    pixels: list[int] = []
                    for begin, end in skypix_bounds[name]:
                        pixels.extend(range(begin, end))
                    bind[f"{name}_pixels"] = pixels
                try:
                    return butler.query_datasets(
                        self.dataset_type_node.dataset_type,
                        collections=input_collections,
                        data_id=data_id.subset(self.constraint_dimensions),
                        where=" AND ".join(where_terms),
                        bind=bind,
                        with_dimension_records=True,
                        limit=None,
                        explain=False,
                    )
                except MissingDatasetTypeError:
                    return []
            else:
                raise NotImplementedError(
                    f"No support for skypix lookup {self.task_node.label}.{self.edge.connection_name} "
                    "that requires additional spatial and/or temporal constraints. "
                    "Please create a feature-request ticket and use a lookup function in the meantime."
                )
        if self._bounds.spatial_connections or self._bounds.temporal_connections:
            raise NotImplementedError(
                f"No support for prerequisite lookup {self.task_node.label}.{self.edge.connection_name} "
                "that requires other connections to determine spatial or temporal bounds but does not "
                "fit into one of our standard cases. "
                "Please create a feature-request ticket and use a lookup function in the meantime."
            )
        # If the spatial/temporal bounds are not customized, and the dataset
        # doesn't have any skypix dimensions, a vanilla query_datasets call
        # should work.  This case should always be optimized by
        # AllDimensionsQuantumGraphBuilder as well.  Note that we use the
        # original quantum data ID here, not those with constraint_dimensions
        # that strips out the spatial/temporal stuff, because here we want the
        # butler query system to handle the spatial/temporal stuff like it
        # normally would.
        try:
            return butler.query_datasets(
                self.dataset_type_node.dataset_type,
                collections=input_collections,
                data_id=data_id,
                with_dimension_records=True,
                limit=None,
                explain=False,
            )
        except MissingDatasetTypeError:
            return []


@dataclasses.dataclass
class PrerequisiteBounds:
    """A QuantumGraph-generation helper class that manages the spatial and
    temporal bounds of a tasks' quanta, for the purpose of finding
    prerequisite inputs.
    """

    task_node: TaskNode
    """The `~pipeline_graph.PipelineGraph` node that represents the task."""

    spatial_connections: frozenset[str] = dataclasses.field(init=False)
    """Regular input or output connections whose (assumed spatial) data IDs
    should be used to define the spatial bounds of this task's quanta.

    See Also
    --------
    PipelineTaskConnections.getSpatialBoundsConnections
    """

    temporal_connections: frozenset[str] = dataclasses.field(init=False)
    """Regular input or output connections whose (assumed temporal) data IDs
    should be used to define the temporal bounds of this task's quanta.

    See Also
    --------
    PipelineTaskConnections.getTemporalBoundsConnections
    """

    all_dataset_skypix: dict[str, SkyPixDimension] = dataclasses.field(default_factory=dict)
    """The union of all `PrerequisiteFinder.dataset_skypix` attributes for all
    (remaining) prerequisite finders for this task.
    """

    any_dataset_has_timespan: bool = dataclasses.field(default=False)
    """Whether any `PrerequisiteFinder.dataset_has_timespan` attribute is true
    for any (remaining) prerequisite finder for this task.
    """

    def __post_init__(self) -> None:
        self.spatial_connections = frozenset(self.task_node.get_spatial_bounds_connections())
        self.temporal_connections = frozenset(self.task_node.get_temporal_bounds_connections())

    def make_skypix_bounds_builder(self, quantum_data_id: DataCoordinate) -> SkyPixBoundsBuilder:
        """Return an object that accumulates the appropriate spatial bounds for
        a quantum.

        Parameters
        ----------
        quantum_data_id : `lsst.daf.butler.DataCoordinate`
            Data ID for this quantum.

        Returns
        -------
        builder : `SkyPixBoundsBuilder`
            Object that accumulates the appropriate spatial bounds for a
            quantum.  If the spatial bounds are not needed, this object will do
            nothing.
        """
        if not self.all_dataset_skypix:
            return _TrivialSkyPixBoundsBuilder()
        if self.spatial_connections:
            return _ConnectionSkyPixBoundsBuilder(
                self.task_node, self.spatial_connections, self.all_dataset_skypix.values(), quantum_data_id
            )
        if self.task_node.dimensions.spatial:
            return _QuantumOnlySkyPixBoundsBuilder(self.all_dataset_skypix.values(), quantum_data_id)
        else:
            return _UnboundedSkyPixBoundsBuilder(self.all_dataset_skypix.values())

    def make_timespan_builder(self, quantum_data_id: DataCoordinate) -> TimespanBuilder:
        """Return an object that accumulates the appropriate timespan for
        a quantum.

        Parameters
        ----------
        quantum_data_id : `lsst.daf.butler.DataCoordinate`
            Data ID for this quantum.

        Returns
        -------
        builder : `TimespanBuilder`
            Object that accumulates the appropriate timespan bounds for a
            quantum.  If a timespan is not needed, this object will do nothing.
        """
        if not self.any_dataset_has_timespan:
            return _TrivialTimespanBuilder()
        if self.temporal_connections:
            return _ConnectionTimespanBuilder(self.task_node, self.temporal_connections, quantum_data_id)
        if self.task_node.dimensions.temporal:
            return _QuantumOnlyTimespanBuilder(quantum_data_id)
        else:
            return _UnboundedTimespanBuilder()


class SkyPixBoundsBuilder(ABC):
    """A base class for objects that accumulate the appropriate spatial bounds
    for a quantum.
    """

    def handle_dataset(self, parent_dataset_type_name: str, data_id: DataCoordinate) -> None:
        """Handle the skeleton graph node for a regular input/output connection
        for this quantum, including its data ID in the bounds if appropriate.

        Parameters
        ----------
        parent_dataset_type_name : `str`
            Name of the dataset type.  Never a component dataset type name.
        data_id : `lsst.daf.butler.DataCoordinate`
            Data ID for the dataset.
        """
        pass

    @abstractmethod
    def finish(self) -> dict[str, RangeSet]:
        """Finish building the spatial bounds and return them.

        Returns
        -------
        bounds : `dict` [ `str`, `lsst.sphgeom.RangeSet` ]
            The spatial bounds of this quantum in various skypix dimensions.
            Keys are skypix dimension names and values are sets of integer
            pixel ID ranges.
        """
        raise NotImplementedError()


class TimespanBuilder(ABC):
    """A base class for objects that accumulate the appropriate timespan
    for a quantum.
    """

    def handle_dataset(self, parent_dataset_type_name: str, data_id: DataCoordinate) -> None:
        """Handle the skeleton graph node for a regular input/output connection
        for this quantum, including its data ID in the bounds if appropriate.

        Parameters
        ----------
        parent_dataset_type_name : `str`
            Name of the dataset type.  Never a component dataset type name.
        data_id : `lsst.daf.butler.DataCoordinate`
            Data ID for the dataset.
        """
        pass

    @abstractmethod
    def finish(self) -> Timespan | None:
        """Finish building the timespan and return it.

        Returns
        -------
        timespan : `lsst.daf.butler.Timespan` or `None`
            The timespan of this quantum, or `None` if it is known to not be
            needed.
        """
        raise NotImplementedError()


class _TrivialSkyPixBoundsBuilder(SkyPixBoundsBuilder):
    """Implementation of `SkyPixBoundsBuilder` for when no skypix bounds are
    needed.
    """

    def finish(self) -> dict[str, RangeSet]:
        return {}


class _TrivialTimespanBuilder(TimespanBuilder):
    """Implementation of `TimespanBuilder` for when no timespan is needed."""

    def finish(self) -> None:
        return None


class _QuantumOnlySkyPixBoundsBuilder(SkyPixBoundsBuilder):
    """Implementation of `SkyPixBoundsBuilder` for when the quantum data IDs
    provide the only relevant spatial regions.
    """

    def __init__(self, dimensions: Iterable[SkyPixDimension], quantum_data_id: DataCoordinate) -> None:
        self._region = quantum_data_id.region
        self._dimensions = dimensions

    def finish(self) -> dict[str, RangeSet]:
        return {
            dimension.name: dimension.pixelization.envelope(self._region) for dimension in self._dimensions
        }


class _QuantumOnlyTimespanBuilder(TimespanBuilder):
    """Implementation of `TimespanBuilder` for when the quantum data IDs
    provide the only relevant timespans.
    """

    def __init__(self, quantum_data_id: DataCoordinate) -> None:
        self._timespan = cast(Timespan, quantum_data_id.timespan)

    def finish(self) -> Timespan:
        return self._timespan


class _UnboundedSkyPixBoundsBuilder(SkyPixBoundsBuilder):
    """Implementation of `SkyPixBoundsBuilder` for when the bounds cover the
    full sky.
    """

    def __init__(self, dimensions: Iterable[SkyPixDimension]):
        self._dimensions = dimensions

    def finish(self) -> dict[str, RangeSet]:
        return {dimension.name: dimension.pixelization.universe() for dimension in self._dimensions}


class _UnboundedTimespanBuilder(TimespanBuilder):
    """Implementation of `TimespanBuilder` for when the timespan covers all
    time.
    """

    def finish(self) -> Timespan:
        return Timespan(None, None)


class _ConnectionSkyPixBoundsBuilder(SkyPixBoundsBuilder):
    """Implementation of `SkyPixBoundsBuilder` for when other input or output
    connections contribute to the spatial bounds.
    """

    def __init__(
        self,
        task_node: TaskNode,
        bounds_connections: frozenset[str],
        dimensions: Iterable[SkyPixDimension],
        quantum_data_id: DataCoordinate,
    ) -> None:
        self._dimensions = dimensions
        self._regions: list[Region] = []
        if task_node.dimensions.spatial:
            self._regions.append(quantum_data_id.region)
        self._dataset_type_names: set[str] = set()
        for connection_name in bounds_connections:
            if edge := task_node.inputs.get(connection_name):
                self._dataset_type_names.add(edge.parent_dataset_type_name)
            else:
                self._dataset_type_names.add(task_node.outputs[connection_name].parent_dataset_type_name)
            # Note that we end up raising if the input is a prerequisite (and
            # hence not in task_node.inputs or task_node.outputs); this
            # justifies the cast in `handle_dataset`.

    def handle_dataset(self, parent_dataset_type_name: str, data_id: DataCoordinate) -> None:
        if parent_dataset_type_name in self._dataset_type_names:
            self._regions.append(data_id.region)

    def finish(self) -> dict[str, RangeSet]:
        result = {}
        for dimension in self._dimensions:
            bounds = RangeSet()
            for region in self._regions:
                bounds |= dimension.pixelization.envelope(region)
            result[dimension.name] = bounds
        return result


class _ConnectionTimespanBuilder(TimespanBuilder):
    """Implementation of `TimespanBuilder` for when other input or output
    connections contribute to the timespan.
    """

    def __init__(
        self,
        task_node: TaskNode,
        bounds_connections: frozenset[str],
        quantum_data_id: DataCoordinate,
    ) -> None:
        timespan = (
            cast(Timespan, quantum_data_id.timespan)
            if task_node.dimensions.temporal
            else Timespan.makeEmpty()
        )
        self._begin_nsec = timespan.nsec[0]
        self._end_nsec = timespan.nsec[1]
        self._dataset_type_names = set()
        for connection_name in bounds_connections:
            if edge := task_node.inputs.get(connection_name):
                self._dataset_type_names.add(edge.parent_dataset_type_name)
            else:
                self._dataset_type_names.add(task_node.outputs[connection_name].parent_dataset_type_name)
            # Note that we end up raising if the input is a prerequisite (and
            # hence not in task_node.inputs or task_node.outputs); this
            # justifies the cast in `handle_dataset`.

    def handle_dataset(self, parent_dataset_type_name: str, data_id: DataCoordinate) -> None:
        if parent_dataset_type_name in self._dataset_type_names:
            nsec = cast(Timespan, data_id.timespan).nsec
            self._begin_nsec = min(self._begin_nsec, nsec[0])
            self._end_nsec = max(self._end_nsec, nsec[1])

    def finish(self) -> Timespan:
        return Timespan(None, None, _nsec=(self._begin_nsec, self._end_nsec))



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("ProgressCounter", "ProgressManager", "make_worker_log")

import logging
import os
import time
from types import TracebackType
from typing import Any, Self

from lsst.utils.logging import TRACE, VERBOSE, LsstLogAdapter, PeriodicLogger, getLogger

from ._config import AggregatorConfig


class ProgressCounter:
    """A progress tracker for an individual aspect of the aggregation process.

    Parameters
    ----------
    parent : `ProgressManager`
        The parent progress manager object.
    description : `str`
        Human-readable description of this aspect.
    unit : `str`
        Unit (in plural form) for the items being counted.
    total : `int`, optional
        Expected total number of items.  May be set later.
    """

    def __init__(self, parent: ProgressManager, description: str, unit: str, total: int | None = None):
        self._parent = parent
        self.total = total
        self._description = description
        self._current = 0
        self._unit = unit
        self._bar: Any = None

    def update(self, n: int) -> None:
        """Report that ``n`` new items have been processed.

        Parameters
        ----------
        n : `int`
            Number of new items processed.
        """
        self._current += n
        if self._parent.interactive:
            if self._bar is None:
                if n == self.total:
                    return
                from tqdm import tqdm

                self._bar = tqdm(desc=self._description, total=self.total, leave=False, unit=f" {self._unit}")
            else:
                self._bar.update(n)
                if self._current == self.total:
                    self._bar.close()
        self._parent._log_status()

    def close(self) -> None:
        """Close the counter, guaranteeing that `update` will not be called
        again.
        """
        if self._bar is not None:
            self._bar.close()
            self._bar = None

    def append_log_terms(self, msg: list[str]) -> None:
        """Append a log message for this counter to a list if it is active.

        Parameters
        ----------
        msg : `list` [ `str` ]
            List of messages to concatenate into a single line and log
            together, to be modified in-place.
        """
        if self.total is not None and self._current > 0 and self._current < self.total:
            msg.append(f"{self._description} ({self._current} of {self.total} {self._unit})")


class ProgressManager:
    """A helper class for the provenance aggregator that handles reporting
    progress to the user.

    This includes both logging (including periodic logging) and optional
    progress bars.

    Parameters
    ----------
    log : `lsst.utils.logging.LsstLogAdapter`
        LSST-customized logger.
    config : `AggregatorConfig`
        Configuration for the aggregator.

    Notes
    -----
    This class is a context manager in order to manage the redirection of
    logging when progress bars for interactive display are in use.  The context
    manager does nothing otherwise.
    """

    def __init__(self, log: LsstLogAdapter, config: AggregatorConfig):
        self.start = time.time()
        self.log = log
        self.config = config
        self._periodic_log = PeriodicLogger(self.log, config.log_status_interval)
        self.scans = ProgressCounter(self, "scanning", "quanta")
        self.writes = ProgressCounter(self, "writing", "quanta")
        self.quantum_ingests = ProgressCounter(self, "ingesting outputs", "quanta")
        self.interactive = config.interactive_status

    def __enter__(self) -> Self:
        if self.interactive:
            from tqdm.contrib.logging import logging_redirect_tqdm

            self._logging_redirect = logging_redirect_tqdm()
            self._logging_redirect.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        if self.interactive:
            self._logging_redirect.__exit__(exc_type, exc_value, traceback)
        return None

    @property
    def elapsed_time(self) -> float:
        """The time in seconds since the start of the aggregator."""
        return time.time() - self.start

    def _log_status(self) -> None:
        """Invoke the periodic logger with the current status."""
        log_terms: list[str] = []
        self.scans.append_log_terms(log_terms)
        self.writes.append_log_terms(log_terms)
        self.quantum_ingests.append_log_terms(log_terms)
        self._periodic_log.log("Status after %0.1fs: %s.", self.elapsed_time, "; ".join(log_terms))


def make_worker_log(name: str, config: AggregatorConfig) -> LsstLogAdapter:
    """Make a logger for a worker.

    Parameters
    ----------
    name : `str`
        Name of the worker, to be used as part of the name for the logger.
    config : `AggregatorConfig`
        Configuration for the aggregator.
    """
    base_log = logging.getLogger(f"lsst.pipe.base.quantum_graph.aggregator.{name}")
    base_log.propagate = False
    log = getLogger(logger=base_log)
    if config.worker_log_dir is not None:
        os.makedirs(config.worker_log_dir, exist_ok=True)
        match config.worker_log_level.upper():
            case "VERBOSE":
                log.setLevel(VERBOSE)
            case "TRACE":
                log.setLevel(TRACE)
            case std:
                log.setLevel(getattr(logging, std))
        handler = logging.FileHandler(os.path.join(config.worker_log_dir, f"{name}.log"))
        handler.setFormatter(
            logging.Formatter("%(levelname)s %(asctime)s.%(msecs)03d %(message)s", "%Y-%m-%dT%H:%M:%S")
        )
        log.addHandler(handler)
    else:
        log.addHandler(logging.NullHandler())
    return log



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("QuantumGraphDotVisualizer", "QuantumGraphMermaidVisualizer", "QuantumGraphVisualizer")

import html
import uuid
from abc import abstractmethod
from typing import IO, ClassVar, Generic, TypeVar

from ..pipeline_graph import NodeType
from ._common import BaseQuantumGraph, BipartiteEdgeInfo, DatasetInfo, QuantumInfo

# We use the old generic syntax in this module because for some reason the new
# one confused Sphinx (or one of its plugins), even though it seems fine with
# it in other places.  We can try again when we're ready to remove types from
# the docstrings of annotated functions, in case that matters.
_G = TypeVar("_G", bound=BaseQuantumGraph, contravariant=True)
_Q = TypeVar("_Q", bound=QuantumInfo, contravariant=True)
_D = TypeVar("_D", bound=DatasetInfo, contravariant=True)


class QuantumGraphVisualizer(Generic[_G, _Q, _D]):  # noqa: UP046
    """A base class for exporting quantum graphs to graph-visualization
    languages.

    Notes
    -----
    This base class is not intended to support implementations for every
    possible visualization language, but it does neatly unify common logic for
    at least GraphViz and Mermaid.
    """

    @abstractmethod
    def render_header(self, qg: _G, is_bipartite: bool) -> str:
        """Return the beginning of a graph visualization.

        Parameters
        ----------
        qg : `.BaseQuantumGraph`
            Quantum graph to visualize.
        is_bipartite : `bool`
            Whether a bipartite graph visualization is being requested.

        Returns
        -------
        rendered : `str`
            String that starts the visualization.  May contain newlines, but
            an additional newline will automatically be added after the string,
            before the first node.
        """
        raise NotImplementedError()

    @abstractmethod
    def render_footer(self, qg: _G, is_bipartite: bool) -> str:
        """Return the ending of a graph visualization.

        Parameters
        ----------
        qg : `.BaseQuantumGraph`
            Quantum graph to visualize.
        is_bipartite : `bool`
            Whether a bipartite graph visualization is being requested.

        Returns
        -------
        rendered : `str`
            String that ends the visualization.  May contain newlines, but
            an additional newline will automatically be added after the string,
            at the end of the file.
        """
        raise NotImplementedError()

    @abstractmethod
    def render_quantum(self, quantum_id: uuid.UUID, data: _Q, is_bipartite: bool) -> str:
        """Return the representation of a quantum in a graph visualization.

        Parameters
        ----------
        quantum_id : `uuid.UUID`
            ID of the quantum.
        data : `.QuantumInfo`
            Mapping with additional information about the quantum.
        is_bipartite : `bool`
            Whether a bipartite graph visualization is being requested.

        Returns
        -------
        rendered : `str`
            String that represents the quantum.  May contain newlines, but an
            additional newline will automatically be added after the string,
            before the next node or edge.
        """
        raise NotImplementedError()

    @abstractmethod
    def render_dataset(self, dataset_id: uuid.UUID, data: _D) -> str:
        """Return the representation of a dataset in a graph visualization.

        Parameters
        ----------
        dataset_id : `uuid.UUID`
            ID of the dataset.
        data : `.DatasetInfo`
            Mapping with additional information about the dataset.

        Returns
        -------
        rendered : `str`
            String that represents the dataset.  May contain newlines, but an
            additional newline will automatically be added after the string,
            before the next node or edge.
        """
        raise NotImplementedError()

    @abstractmethod
    def render_edge(
        self,
        a: uuid.UUID,
        b: uuid.UUID,
        data: BipartiteEdgeInfo | None,
    ) -> str:
        """Return the representation of an edge a graph visualization.

        Parameters
        ----------
        a : `uuid.UUID`
            ID of the quantum or dataset for which this is an outgoing edge.
        b : `uuid.UUID`
            ID of the quantum or dataset for which this is an incoming edge.
        data : `.BipartiteEdgeInfo`
            Mapping with additional information about the dataset.

        Returns
        -------
        rendered : `str`
            String that represents the edge.  May contain newlines, but an
            additional newline will automatically be added after the string,
            before the next edge.
        """
        raise NotImplementedError()

    def write_quantum_only(self, qg: _G, stream: IO[str]) -> None:
        """Write a visualization for graph with only quantum nodes.

        Parameters
        ----------
        qg : `BaseQuantumGraph`
            Quantum graph to visualize.
        stream : `typing.IO`
            File-like object to write to.
        """
        print(self.render_header(qg, is_bipartite=False), file=stream)
        xgraph = qg.quantum_only_xgraph
        for node_id, node_data in xgraph.nodes.items():
            print(self.render_quantum(node_id, node_data, is_bipartite=False), file=stream)
        for a, b in xgraph.edges:
            print(self.render_edge(a, b, data=None), file=stream)
        print(self.render_footer(qg, is_bipartite=False), file=stream)

    def write_bipartite(self, qg: _G, stream: IO[str]) -> None:
        """Write a visualization for graph with both quantum and dataset nodes.

        Parameters
        ----------
        qg : `BaseQuantumGraph`
            Quantum graph to visualize.
        stream : `typing.IO`
            File-like object to write to.
        """
        print(self.render_header(qg, is_bipartite=True), file=stream)
        xgraph = qg.bipartite_xgraph
        for node_id, node_data in xgraph.nodes.items():
            match node_data["pipeline_node"].key.node_type:
                case NodeType.TASK:
                    print(self.render_quantum(node_id, node_data, is_bipartite=True), file=stream)
                case NodeType.DATASET_TYPE:
                    print(self.render_dataset(node_id, node_data), file=stream)
        for a, b, edge_data in xgraph.edges(data=True):
            print(self.render_edge(a, b, edge_data), file=stream)
        print(self.render_footer(qg, is_bipartite=True), file=stream)


class QuantumGraphDotVisualizer(QuantumGraphVisualizer[BaseQuantumGraph, QuantumInfo, DatasetInfo]):
    """A visualizer for quantum graphs in the GraphViz dot language."""

    def render_header(self, qg: BaseQuantumGraph, is_bipartite: bool) -> str:
        return "\n".join(
            [
                "digraph QuantumGraph {",
                self._render_default("graph", self._ATTRIBS["defaultGraph"]),
                self._render_default("node", self._ATTRIBS["defaultNode"]),
                self._render_default("edge", self._ATTRIBS["defaultEdge"]),
            ]
        )

    def render_footer(self, qg: BaseQuantumGraph, is_bipartite: bool) -> str:
        return "}"

    def render_quantum(self, quantum_id: uuid.UUID, data: QuantumInfo, is_bipartite: bool) -> str:
        labels = [f"{quantum_id}", html.escape(data["task_label"])]
        data_id = data["data_id"]
        labels.extend(f"{key} = {value}" for key, value in data_id.required.items())
        return self._render_node(self._make_node_id(quantum_id), "quantum", labels)

    def render_dataset(self, dataset_id: uuid.UUID, data: DatasetInfo) -> str:
        labels = [html.escape(data["dataset_type_name"]), f"run: {data['run']!r}"]
        data_id = data["data_id"]
        labels.extend(f"{key} = {value}" for key, value in data_id.required.items())
        return self._render_node(self._make_node_id(dataset_id), "dataset", labels)

    def render_edge(self, a: uuid.UUID, b: uuid.UUID, data: BipartiteEdgeInfo | None) -> str:
        return f'"{self._make_node_id(a)}" -> "{self._make_node_id(b)}";'

    _ATTRIBS: ClassVar = dict(
        defaultGraph=dict(splines="ortho", nodesep="0.5", ranksep="0.75", pad="0.5"),
        defaultNode=dict(shape="box", fontname="Monospace", fontsize="14", margin="0.2,0.1", penwidth="3"),
        defaultEdge=dict(color="black", arrowsize="1.5", penwidth="1.5"),
        task=dict(style="filled", color="black", fillcolor="#B1F2EF"),
        quantum=dict(style="filled", color="black", fillcolor="#B1F2EF"),
        dsType=dict(style="rounded,filled,bold", color="#00BABC", fillcolor="#F5F5F5"),
        dataset=dict(style="rounded,filled,bold", color="#00BABC", fillcolor="#F5F5F5"),
    )

    def _render_default(self, type: str, attribs: dict[str, str]) -> str:
        """Set default attributes for a given type."""
        default_attribs = ", ".join([f'{key}="{val}"' for key, val in attribs.items()])
        return f"{type} [{default_attribs}];"

    def _render_node(self, name: str, style: str, labels: list[str]) -> str:
        """Render GV node"""
        label = r"</TD></TR><TR><TD>".join(labels)
        attrib_dict = dict(self._ATTRIBS[style], label=label)
        pre = '<<TABLE BORDER="0" CELLPADDING="5"><TR><TD>'
        post = "</TD></TR></TABLE>>"
        attrib = ", ".join(
            [
                f'{key}="{val}"' if key != "label" else f"{key}={pre}{val}{post}"
                for key, val in attrib_dict.items()
            ]
        )
        return f'"{name}" [{attrib}];'

    def _make_node_id(self, node_id: uuid.UUID) -> str:
        """Return a GV node ID from a quantum or dataset UUID."""
        return f"u{node_id.hex}"


class QuantumGraphMermaidVisualizer(QuantumGraphVisualizer[BaseQuantumGraph, QuantumInfo, DatasetInfo]):
    """A visualizer for quantum graphs in the Mermaid language."""

    def render_header(self, qg: BaseQuantumGraph, is_bipartite: bool) -> str:
        return "flowchart TD"

    def render_footer(self, qg: BaseQuantumGraph, is_bipartite: bool) -> str:
        return ""

    def render_quantum(self, quantum_id: uuid.UUID, data: QuantumInfo, is_bipartite: bool) -> str:
        label_lines = [f"**{data['task_label']}**", f"ID: {quantum_id}"]
        for k, v in data["data_id"].required.items():
            label_lines.append(f"{k}={v}")
        label = "<br>".join(label_lines)
        return f'{self._make_node_id(quantum_id)}["{label}"]'

    def render_dataset(self, dataset_id: uuid.UUID, data: DatasetInfo) -> str:
        label_lines = [
            f"**{data['dataset_type_name']}**",
            f"ID: {dataset_id}",
            f"run: {data['run']}",
        ]
        for k, v in data["data_id"].required.items():
            label_lines.append(f"{k}={v}")
        label = "<br>".join(label_lines)
        return f'{self._make_node_id(dataset_id)}["{label}"]'

    def render_edge(self, a: uuid.UUID, b: uuid.UUID, data: BipartiteEdgeInfo | None) -> str:
        return f"{self._make_node_id(a)} --> {self._make_node_id(b)}"

    def _make_node_id(self, node_id: uuid.UUID) -> str:
        return f"u{node_id.hex}"



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["ExecutionGraphFixup"]

import itertools
import uuid
from collections import defaultdict
from collections.abc import Mapping, Sequence

import networkx as nx

from lsst.daf.butler import DataCoordinate, DataIdValue

from .execution_graph_fixup import ExecutionGraphFixup
from .graph import QuantumGraph


class ExecFixupDataId(ExecutionGraphFixup):
    """Implementation of ExecutionGraphFixup for ordering of tasks based
    on DataId values.

    This class is a trivial implementation mostly useful as an example,
    though it can be used to make actual fixup instances by defining
    a method that instantiates it, e.g.::

        # lsst/ap/verify/ci_fixup.py

        from lsst.pipe.base.exec_fixup_data_id import ExecFixupDataId


        def assoc_fixup():
            return ExecFixupDataId(
                taskLabel="ap_assoc", dimensions=("visit", "detector")
            )

    and then executing pipetask::

        pipetask run --graph-fixup=lsst.ap.verify.ci_fixup.assoc_fixup ...

    This will add new dependencies between quanta executed by the task with
    label "ap_assoc". Quanta with higher visit number will depend on quanta
    with lower visit number and their execution will wait until lower visit
    number finishes.

    Parameters
    ----------
    taskLabel : `str`
        The label of the task for which to add dependencies.
    dimensions : `str` or sequence [`str`]
        One or more dimension names, quanta execution will be ordered
        according to values of these dimensions.
    reverse : `bool`, optional
        If `False` (default) then quanta with higher values of dimensions
        will be executed after quanta with lower values, otherwise the order
        is reversed.
    """

    def __init__(self, taskLabel: str, dimensions: str | Sequence[str], reverse: bool = False):
        self.taskLabel = taskLabel
        self.dimensions = dimensions
        self.reverse = reverse
        if isinstance(self.dimensions, str):
            self.dimensions = (self.dimensions,)
        else:
            self.dimensions = tuple(self.dimensions)

    def fixupQuanta(self, graph: QuantumGraph) -> QuantumGraph:
        raise NotImplementedError()

    def fixup_graph(
        self, xgraph: nx.DiGraph, quanta_by_task: Mapping[str, Mapping[DataCoordinate, uuid.UUID]]
    ) -> None:
        quanta_by_sort_key: defaultdict[tuple[DataIdValue, ...], list[uuid.UUID]] = defaultdict(list)
        for data_id, quantum_id in quanta_by_task[self.taskLabel].items():
            key = tuple(data_id[dim] for dim in self.dimensions)
            quanta_by_sort_key[key].append(quantum_id)
        sorted_keys = sorted(quanta_by_sort_key.keys(), reverse=self.reverse)
        for prev_key, key in itertools.pairwise(sorted_keys):
            xgraph.add_edges_from(itertools.product(quanta_by_sort_key[prev_key], quanta_by_sort_key[key]))



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import operator
from collections.abc import Iterable
from typing import Any

import click

from lsst.daf.butler.cli.opt import (
    dataset_type_option,
    options_file_option,
    register_dataset_types_option,
    repo_argument,
    transfer_dimensions_option,
    transfer_option,
)
from lsst.daf.butler.cli.utils import ButlerCommand, split_commas, unwrap

from ... import script
from ..._status import QuantumAttemptStatus, QuantumSuccessCaveats
from ...quantum_graph import aggregator
from ..opt import instrument_argument, update_output_chain_option


@click.command(short_help="Add an instrument definition to the repository", cls=ButlerCommand)
@repo_argument(required=True)
@instrument_argument(required=True, nargs=-1, help="The fully-qualified name of an Instrument subclass.")
@click.option("--update", is_flag=True)
def register_instrument(*args: Any, **kwargs: Any) -> None:
    """Add an instrument to the data repository."""
    script.register_instrument(*args, **kwargs)


@click.command(short_help="Transfer datasets from a graph to a butler.", cls=ButlerCommand)
@click.argument("graph", required=True)
@click.argument("dest", required=True)
@register_dataset_types_option()
@transfer_dimensions_option(default=False)
@update_output_chain_option()
@click.option(
    "--dry-run", is_flag=True, default=False, help="Run the transfer but do not update the destination butler"
)
@dataset_type_option(help="Subset of dataset types to transfer from graph.")
@options_file_option()
def transfer_from_graph(**kwargs: Any) -> None:
    """Transfer datasets from a quantum graph to a destination butler.

    GRAPH is a URI to the source quantum graph file.

    DEST is a URI to the Butler repository that will receive copies of the
    datasets.
    """
    number = script.transfer_from_graph(**kwargs)
    print(f"Number of datasets transferred: {number}")


@click.command(short_help="Make Zip archive from output files using graph.", cls=ButlerCommand)
@click.argument("graph", required=True)
@repo_argument(
    required=True,
    help="REPO is a URI to a butler configuration that is used to configure "
    "the datastore of the quantum-backed butler.",
)
@click.argument("dest", required=True)
@dataset_type_option(help="Dataset types to include in Zip archive.")
@options_file_option()
def zip_from_graph(**kwargs: Any) -> None:
    """Transfer datasets from a quantum graph to a Zip archive.

    GRAPH is a URI to the source quantum graph file to use when building the
    Zip archive.

    DEST is a directory to write the Zip archive.
    """
    zip = script.zip_from_graph(**kwargs)
    print(f"Zip archive written to {zip}")


@click.command(short_help="Retrieve artifacts from subset of graph.", cls=ButlerCommand)
@click.argument("graph", required=True)
@repo_argument(
    required=True,
    help="REPO is a URI to a butler configuration that is used to configure "
    "the datastore of the quantum-backed butler.",
)
@click.argument("dest", required=True)
@transfer_option()
@click.option(
    "--preserve-path/--no-preserve-path",
    is_flag=True,
    default=True,
    help="Preserve the datastore path to the artifact at the destination.",
)
@click.option(
    "--clobber/--no-clobber",
    is_flag=True,
    default=False,
    help="If clobber, overwrite files if they exist locally.",
)
@click.option(
    "--qgraph-node-id",
    callback=split_commas,
    multiple=True,
    help=unwrap(
        """Only load a specified set of nodes when graph is
        loaded from a file, nodes are identified by UUID
        values. One or more comma-separated strings are
        accepted. By default all nodes are loaded. Ignored if
        graph is not loaded from a file."""
    ),
)
@click.option(
    "--include-inputs/--no-include-inputs",
    is_flag=True,
    default=True,
    help="Whether to include input datasets in retrieval.",
)
@click.option(
    "--include-outputs/--no-include-outputs",
    is_flag=True,
    default=True,
    help="Whether to include output datasets in retrieval.",
)
@options_file_option()
def retrieve_artifacts_for_quanta(**kwargs: Any) -> None:
    """Retrieve artifacts from given quanta defined in quantum graph.

    GRAPH is a URI to the source quantum graph file to use when building the
    Zip archive.

    DEST is a directory to write the Zip archive.
    """
    artifacts = script.retrieve_artifacts_for_quanta(**kwargs)
    print(f"Written {len(artifacts)} artifacts to {kwargs['dest']}.")


_AGGREGATOR_DEFAULTS = aggregator.AggregatorConfig()


@click.command(short_help="Scan for the outputs of an active or completed quantum graph.", cls=ButlerCommand)
@click.argument("predicted_graph", required=True)
@repo_argument(required=True, help="Path or alias for the butler repository.")
@click.option(
    "-o",
    "--output",
    "output_path",
    default=_AGGREGATOR_DEFAULTS.output_path,
    help=(
        "Path to the output provenance quantum graph.  THIS OPTION IS FOR "
        "DEVELOPMENT AND DEBUGGING ONLY.  IT MAY BE REMOVED IN THE FUTURE."
    ),
)
@click.option(
    "--processes",
    "-j",
    "n_processes",
    default=_AGGREGATOR_DEFAULTS.n_processes,
    type=click.IntRange(min=1),
    help="Number of processes to use.",
)
@click.option(
    "--incomplete/--complete",
    "incomplete",
    default=_AGGREGATOR_DEFAULTS.incomplete,
    help="Whether execution has completed (and failures cannot be retried).",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=_AGGREGATOR_DEFAULTS.dry_run,
    help="Do not actually perform any central database ingests.",
)
@click.option(
    "--interactive-status/--no-interactive-status",
    "interactive_status",
    default=_AGGREGATOR_DEFAULTS.interactive_status,
    help="Use progress bars for status reporting instead of periodic logging.",
)
@click.option(
    "--log-status-interval",
    type=int,
    default=_AGGREGATOR_DEFAULTS.log_status_interval,
    help="Interval (in seconds) between periodic logger status updates.",
)
@click.option(
    "--register-dataset-types/--no-register-dataset-types",
    default=_AGGREGATOR_DEFAULTS.register_dataset_types,
    help="Register output dataset types.",
)
@click.option(
    "--update-output-chain/--no-update-output-chain",
    default=_AGGREGATOR_DEFAULTS.update_output_chain,
    help="Prepend the output RUN collection to the output CHAINED collection.",
)
@click.option(
    "--worker-log-dir",
    type=str,
    default=_AGGREGATOR_DEFAULTS.worker_log_dir,
    help="Path to a directory (POSIX only) for parallel worker logs.",
)
@click.option(
    "--worker-log-level",
    type=str,
    default=_AGGREGATOR_DEFAULTS.worker_log_level,
    help="Log level for worker processes/threads (use DEBUG for per-quantum messages).",
)
@click.option(
    "--zstd-level",
    type=int,
    default=_AGGREGATOR_DEFAULTS.zstd_level,
    help="Compression level for the provenance quantum graph file.",
)
@click.option(
    "--zstd-dict-size",
    type=int,
    default=_AGGREGATOR_DEFAULTS.zstd_dict_size,
    help="Size (in bytes) of the ZStandard compression dictionary.",
)
@click.option(
    "--zstd-dict-n-inputs",
    type=int,
    default=_AGGREGATOR_DEFAULTS.zstd_dict_n_inputs,
    help=("Number of samples of each type to include in ZStandard compression dictionary training."),
)
@click.option(
    "--mock-storage-classes/--no-mock-storage-classes",
    default=_AGGREGATOR_DEFAULTS.mock_storage_classes,
    help="Enable support for storage classes created by the lsst.pipe.base.tests.mocks package.",
)
@click.option(
    "--promise-ingest-graph/--no-promise-ingest-graph",
    default=_AGGREGATOR_DEFAULTS.promise_ingest_graph,
    help=(
        "Promise to run 'butler ingest-graph' later, allowing aggregate-graph "
        "to skip metadata/log/config ingestion for now."
    ),
)
def aggregate_graph(predicted_graph: str, repo: str, **kwargs: Any) -> None:
    """Scan for quantum graph's outputs to gather provenance, ingest datasets
    into the central butler repository, and delete datasets that are no
    longer needed.
    """
    # It'd be nice to allow to the user to provide a path to an
    # AggregatorConfig JSON file for options that weren't provided, but Click
    # 8.1 fundamentally cannot handle flag options that default to None rather
    # than True or False (i.e. so they fall back to the config value when not
    # set).  It's not clear whether Click 8.2.x has actually fixed this; Click
    # 8.2.0 tried but caused new problems.

    config = aggregator.AggregatorConfig(**kwargs)
    try:
        aggregator.aggregate_graph(predicted_graph, repo, config)
    except aggregator.FatalWorkerError as err:
        # When this exception is raised, we'll have already logged the relevant
        # traceback from a separate worker.
        raise click.ClickException(str(err)) from None


@click.command(
    short_help="Ingest a provenance quantum graph into a butler.",
    cls=ButlerCommand,
)
@repo_argument(required=True, help="Path or alias for the butler repository.")
@click.argument("provenance_graph", required=False)
@transfer_option(default="move")
@click.option("--batch-size", default=10000, help="How many datasets to process in each transaction.")
@click.option(
    "--output-run",
    default=None,
    help=(
        "Name of the output RUN collection.  Must be provided if the provenance graph is not"
        " provided (so the graph can be found in the butler)."
    ),
)
def ingest_graph(
    *,
    repo: str,
    provenance_graph: str | None,
    transfer: str | None,
    batch_size: int,
    output_run: str | None,
) -> None:
    """Ingest a provenance graph into a butler repository."""
    from ...quantum_graph.ingest_graph import ingest_graph as ingest_graph_py

    ingest_graph_py(repo, provenance_graph, transfer=transfer, batch_size=batch_size, output_run=output_run)


@click.command(
    short_help="Print and write provenance reports.",
    cls=ButlerCommand,
)
@click.argument("repo_or_qg")
@click.argument("collection", required=False, default=None)
@click.option(
    "--state",
    multiple=True,
    type=click.Choice(QuantumAttemptStatus),
    help=(
        "Additional quantum state to include in the status report and data ID tables "
        "(FAILED, ABORTED, and ABORTED_SUCCESS are included by default)."
    ),
)
@click.option(
    "--no-state",
    multiple=True,
    type=str,
    metavar="STATE",
    help="Quantum state to drop from in status report and data ID tables (same options as --state).",
)
@click.option(
    "--status-report",
    default=None,
    metavar="URI",
    help="File or URI (.json) for a detailed report (with data IDs) on quanta with certain states.",
)
@click.option(
    "--quantum-table/--no-quantum-table",
    default=True,
    help="Whether to print summary of quantum status counts to STDOUT.",
)
@click.option(
    "--exception-table/--no-exception-table",
    default=True,
    help="Whether to print summary of exception type counts STDOUT.",
)
@click.option(
    "--caveat",
    multiple=True,
    type=click.Choice(QuantumSuccessCaveats),
    help=(
        "Include successful quanta in the status report if they have this caveat. "
        "May be passed multiple times; any matching caveat is included. "
        "Passing this option implicitly adds '--state SUCCESSFUL'."
    ),
)
@click.option(
    "--data-id-table-dir",
    default=None,
    metavar="URI",
    help=(
        "Directory (may be a URI) for a tree of data ID tables for each "
        "task label, status, and exception type combination in the status report."
    ),
)
def provenance_report(
    *,
    repo_or_qg: str,
    collection: str | None,
    state: Iterable[QuantumAttemptStatus],
    no_state: Iterable[str],
    status_report: str | None,
    quantum_table: bool = False,
    exception_table: bool = False,
    caveat: Iterable[QuantumSuccessCaveats],
    data_id_table_dir: str | None,
) -> None:
    """Read a provenance quantum graph from a butler or file and use it to
    generate reports.

    REPO_OR_QG is a path or alias for the butler repository (if reading an
    ingested graph, as indicated by passing COLLECTION), or the path to a
    provenance quantum graph file.
    """
    from ...quantum_graph import ProvenanceQuantumGraph

    states = set(state)
    states.add(QuantumAttemptStatus.FAILED)
    states.add(QuantumAttemptStatus.ABORTED)
    states.add(QuantumAttemptStatus.ABORTED_SUCCESS)
    for state_name in no_state:
        states.discard(QuantumAttemptStatus.__members__[state_name])
    with_caveats: QuantumSuccessCaveats | None = None
    if caveat:
        states.add(QuantumAttemptStatus.SUCCESSFUL)
        with_caveats = functools.reduce(
            operator.__or__,
            caveat,
            QuantumSuccessCaveats.NO_CAVEATS,
        )
    with ProvenanceQuantumGraph.from_args(repo_or_qg, collection=collection, datasets=()) as (graph, _):
        graph.make_many_reports(
            status_report_file=status_report,
            states=states,
            print_quantum_table=quantum_table,
            print_exception_table=exception_table,
            with_caveats=with_caveats,
            data_id_table_dir=data_id_table_dir,
        )



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The base class for the QuantumGraph-generation algorithm and various
helper classes.
"""

from __future__ import annotations

__all__ = (
    "EmptyDimensionsDatasets",
    "OutputExistsError",
    "PrerequisiteMissingError",
    "QuantumGraphBuilder",
    "QuantumGraphBuilderError",
)

import dataclasses
import operator
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, cast, final

from lsst.daf.butler import (
    Butler,
    CollectionType,
    DataCoordinate,
    DatasetRef,
    DatasetType,
    DimensionDataAttacher,
    DimensionUniverse,
    NamedKeyDict,
    NamedKeyMapping,
    Quantum,
)
from lsst.daf.butler._rubin import generate_uuidv7
from lsst.daf.butler.datastore.record_data import DatastoreRecordData
from lsst.daf.butler.registry import MissingCollectionError, MissingDatasetTypeError
from lsst.utils.logging import LsstLogAdapter, getLogger
from lsst.utils.timer import timeMethod

from . import automatic_connection_constants as acc
from ._status import NoWorkFound
from ._task_metadata import TaskMetadata
from .connections import AdjustQuantumHelper, QuantaAdjuster
from .pipeline_graph import Edge, PipelineGraph, TaskNode
from .prerequisite_helpers import PrerequisiteInfo, SkyPixBoundsBuilder, TimespanBuilder
from .quantum_graph_skeleton import (
    DatasetKey,
    PrerequisiteDatasetKey,
    QuantumGraphSkeleton,
    QuantumKey,
    TaskInitKey,
)

if TYPE_CHECKING:
    from .graph import QuantumGraph
    from .pipeline import TaskDef
    from .quantum_graph import PredictedDatasetModel, PredictedQuantumGraphComponents


class QuantumGraphBuilderError(Exception):
    """Base class for exceptions generated by QuantumGraphBuilder."""

    pass


class OutputExistsError(QuantumGraphBuilderError):
    """Exception generated when output datasets already exist."""

    pass


class PrerequisiteMissingError(QuantumGraphBuilderError):
    """Exception generated when a prerequisite dataset does not exist."""

    pass


class InitInputMissingError(QuantumGraphBuilderError):
    """Exception generated when an init-input dataset does not exist."""

    pass


class QuantumGraphBuilder(ABC):
    """An abstract base class for building `.QuantumGraph` objects from a
    pipeline.

    Parameters
    ----------
    pipeline_graph : `.pipeline_graph.PipelineGraph`
        Pipeline to build a `.QuantumGraph` from, as a graph.  Will be resolved
        in-place with the given butler (any existing resolution is ignored).
    butler : `lsst.daf.butler.Butler`
        Client for the data repository.  Should be read-only.
    input_collections : `~collections.abc.Sequence` [ `str` ], optional
        Collections to search for overall-input datasets.  If not provided,
        ``butler.collections`` is used (and must not be empty).
    output_run : `str`, optional
        Output `~lsst.daf.butler.CollectionType.RUN` collection.  If not
        provided, ``butler.run`` is used (and must not be `None`).
    skip_existing_in : `~collections.abc.Sequence` [ `str` ], optional
        Collections to search for outputs that already exist for the purpose of
        skipping quanta that have already been run.
    clobber : `bool`, optional
        Whether to raise if predicted outputs already exist in ``output_run``
        (not including those quanta that would be skipped because they've
        already been run).  This never actually clobbers outputs; it just
        informs the graph generation algorithm whether execution will run with
        clobbering enabled.  This is ignored if ``output_run`` does not exist.

    Notes
    -----
    Constructing a `QuantumGraphBuilder` will run queries for existing datasets
    with empty data IDs (including but not limited to init inputs and outputs),
    in addition to resolving the given pipeline graph and testing for existence
    of the ``output`` run collection.

    The `build` method splits the pipeline graph into independent subgraphs,
    then calls the abstract method `process_subgraph` on each, to allow
    concrete implementations to populate the rough graph structure (the
    `~.quantum_graph_skeleton.QuantumGraphSkeleton` class), including searching
    for existing datasets.  The `build` method then:

    - assembles `lsst.daf.butler.Quantum` instances from all data IDs in the
      skeleton;
    - looks for existing outputs found in ``skip_existing_in`` to see if any
      quanta should be skipped;
    - calls `PipelineTaskConnections.adjustQuantum` on all quanta, adjusting
      downstream quanta appropriately when preliminary predicted outputs are
      rejected (pruning nodes that will not have the inputs they need to run);
    - attaches datastore records and registry dataset types to the graph.

    In addition to implementing `process_subgraph`, derived classes are
    generally expected to add new construction keyword-only arguments to
    control the data IDs of the quantum graph, while forwarding all of the
    arguments defined in the base class to `super`.
    """

    def __init__(
        self,
        pipeline_graph: PipelineGraph,
        butler: Butler,
        *,
        input_collections: Sequence[str] | None = None,
        output_run: str | None = None,
        skip_existing_in: Sequence[str] = (),
        clobber: bool = False,
    ):
        self.log = getLogger(__name__)
        self.metadata = TaskMetadata()
        self._pipeline_graph = pipeline_graph
        if input_collections is None:
            input_collections = butler.collections.defaults
        if not input_collections:
            raise ValueError("No input collections provided.")
        self.input_collections = input_collections
        if output_run is None:
            output_run = butler.run
        if not output_run:
            raise ValueError("No output RUN collection provided.")
        self.butler = butler.clone(collections=input_collections)
        self.output_run = output_run
        self.skip_existing_in = skip_existing_in
        self.empty_data_id = DataCoordinate.make_empty(butler.dimensions)
        self.clobber = clobber
        # See whether the output run already exists.
        self.output_run_exists = False
        try:
            if self.butler.registry.getCollectionType(self.output_run) is not CollectionType.RUN:
                raise RuntimeError(f"{self.output_run!r} is not a RUN collection.")
            self.output_run_exists = True
        except MissingCollectionError:
            # If the run doesn't exist we never need to clobber.  This is not
            # an error so you can run with clobber=True the first time you
            # attempt some processing as well as all subsequent times, instead
            # of forcing the user to make the first attempt different.
            self.clobber = False
        # We need to know whether the skip_existing_in collection sequence
        # starts with the output run collection, as an optimization to avoid
        # queries later.
        try:
            skip_existing_in_flat = self.butler.collections.query(self.skip_existing_in, flatten_chains=True)
        except MissingCollectionError:
            skip_existing_in_flat = []
        if not skip_existing_in_flat:
            self.skip_existing_in = []
        if self.skip_existing_in and self.output_run_exists:
            self.skip_existing_starts_with_output_run = self.output_run == skip_existing_in_flat[0]
        else:
            self.skip_existing_starts_with_output_run = False
        try:
            packages_storage_class = butler.get_dataset_type(acc.PACKAGES_INIT_OUTPUT_NAME).storageClass_name
        except MissingDatasetTypeError:
            packages_storage_class = acc.PACKAGES_INIT_OUTPUT_STORAGE_CLASS
        self._global_init_output_types = {
            acc.PACKAGES_INIT_OUTPUT_NAME: DatasetType(
                acc.PACKAGES_INIT_OUTPUT_NAME,
                self.universe.empty,
                packages_storage_class,
            )
        }
        with self.butler.registry.caching_context():
            self._pipeline_graph.resolve(self.butler.registry)
            self.empty_dimensions_datasets = self._find_empty_dimension_datasets()
            self.prerequisite_info = {
                task_node.label: PrerequisiteInfo(task_node, self._pipeline_graph)
                for task_node in pipeline_graph.tasks.values()
            }

    log: LsstLogAdapter
    """Logger to use for all quantum-graph generation messages.

    General and per-task status messages should be logged at `~logging.INFO`
    level or higher, per-dataset-type status messages should be logged at
    `~lsst.utils.logging.VERBOSE` or higher, and per-data-ID status messages
    should be logged at `logging.DEBUG` or higher.
    """

    metadata: TaskMetadata
    """Metadata to store in the QuantumGraph.

    The `TaskMetadata` class is used here primarily in order to enable
    resource-usage collection with the `lsst.utils.timer.timeMethod` decorator.
    """

    butler: Butler
    """Client for the data repository.

    Should be read-only.
    """

    input_collections: Sequence[str]
    """Collections to search for overall-input datasets.
    """

    output_run: str
    """Output `~lsst.daf.butler.CollectionType.RUN` collection.
    """

    skip_existing_in: Sequence[str]
    """Collections to search for outputs that already exist for the purpose
    of skipping quanta that have already been run.
    """

    clobber: bool
    """Whether to raise if predicted outputs already exist in ``output_run``

    This never actually clobbers outputs; it just informs the graph generation
    algorithm whether execution will run with clobbering enabled.  This is
    always `False` if `output_run_exists` is `False`.
    """

    empty_data_id: DataCoordinate
    """An empty data ID in the data repository's dimension universe.
    """

    output_run_exists: bool
    """Whether the output run exists in the data repository already.
    """

    skip_existing_starts_with_output_run: bool
    """Whether the `skip_existing_in` sequence begins with `output_run`.

    If this is true, any dataset found in `output_run` can be used to
    short-circuit queries in `skip_existing_in`.
    """

    empty_dimensions_datasets: EmptyDimensionsDatasets
    """Struct holding datasets with empty dimensions that have already been
    found in the data repository.
    """

    prerequisite_info: Mapping[str, PrerequisiteInfo]
    """Helper objects for finding prerequisite inputs, organized by task label.

    Subclasses that find prerequisites should remove the
    covered `~prerequisite_helpers.PrerequisiteFinder` objects from this
    attribute.
    """

    @property
    def universe(self) -> DimensionUniverse:
        """Definitions of all data dimensions."""
        return self.butler.dimensions

    @final
    @timeMethod
    def build(
        self, metadata: Mapping[str, Any] | None = None, attach_datastore_records: bool = True
    ) -> QuantumGraph:
        """Build the quantum graph, returning an old `QuantumGraph` instance.

        Parameters
        ----------
        metadata : `~collections.abc.Mapping`, optional
            Flexible metadata to add to the quantum graph.
        attach_datastore_records : `bool`, optional
            Whether to include datastore records in the graph.  Required for
            `lsst.daf.butler.QuantumBackedButler` execution.

        Returns
        -------
        quantum_graph : `.QuantumGraph`
            DAG describing processing to be performed.

        Notes
        -----
        External code is expected to construct a `QuantumGraphBuilder` and then
        call this method exactly once.  See class documentation for details on
        what it does.
        """
        skeleton = self._build_skeleton(attach_datastore_records=attach_datastore_records)
        if metadata is None:
            metadata = {
                "input": list(self.input_collections),
                "output_run": self.output_run,
            }
        return self._construct_quantum_graph(skeleton, metadata)

    def finish(
        self,
        output: str | None = None,
        metadata: Mapping[str, Any] | None = None,
        attach_datastore_records: bool = True,
    ) -> PredictedQuantumGraphComponents:
        """Return quantum graph components that can be used to save or
        construct a `PredictedQuantumGraph` instance.

        Parameters
        ----------
        output : `str` or `None`, optional
            Output `~lsst.daf.butler.CollectionType.CHAINED` collection that
            combines the input and output collections.
        metadata : `~collections.abc.Mapping`, optional
            Mapping of JSON-friendly metadata.  Collection information, the
            current user, and the current timestamp are automatically
            included.
        attach_datastore_records : `bool`, optional
            Whether to include datastore records for overall inputs for
            `~lsst.daf.butler.QuantumBackedButler`.

        Returns
        -------
        components : `.quantum_graph.PredictedQuantumGraphComponents`
            Components that can be used to construct a graph object and/or save
            it to disk.
        """
        skeleton = self._build_skeleton(attach_datastore_records=attach_datastore_records)
        return self._construct_components(skeleton, output=output, metadata=metadata)

    def _build_skeleton(self, attach_datastore_records: bool = True) -> QuantumGraphSkeleton:
        """Build a complete skeleton for the quantum graph.

        Parameters
        ----------
        attach_datastore_records : `bool`, optional
            Whether to include datastore records in the graph.  Required for
            `lsst.daf.butler.QuantumBackedButler` execution.

        Returns
        -------
        quantum_graph_skeleton : `QuantumGraphSkeleton`
            DAG describing processing to be performed.
        """
        with self.butler.registry.caching_context():
            full_skeleton = QuantumGraphSkeleton(self._pipeline_graph.tasks)
            subgraphs = list(self._pipeline_graph.split_independent())
            for i, subgraph in enumerate(subgraphs):
                self.log.info(
                    "Processing pipeline subgraph %d of %d with %d task(s).",
                    i + 1,
                    len(subgraphs),
                    len(subgraph.tasks),
                )
                self.log.verbose("Subgraph tasks: [%s]", ", ".join(label for label in subgraph.tasks))
                subgraph_skeleton = self.process_subgraph(subgraph)
                full_skeleton.update(subgraph_skeleton)
            # Loop over tasks to apply skip-existing logic and add missing
            # prerequisites.  The pipeline graph must be topologically sorted,
            # so a quantum is only processed after any quantum that provides
            # its inputs has been processed.
            skipped_quanta: dict[str, list[QuantumKey]] = {}
            for task_node in self._pipeline_graph.tasks.values():
                skipped_quanta[task_node.label] = self._resolve_task_quanta(task_node, full_skeleton)
            # Add any dimension records not handled by the subclass, and
            # aggregate any that were added directly to data IDs.
            full_skeleton.attach_dimension_records(self.butler, self._pipeline_graph.get_all_dimensions())
            # Loop over tasks again to run the adjust hooks.
            for task_node in self._pipeline_graph.tasks.values():
                self._adjust_task_quanta(task_node, full_skeleton, skipped_quanta[task_node.label])
            # Add global init-outputs to the skeleton.
            for dataset_type in self._global_init_output_types.values():
                dataset_key = full_skeleton.add_dataset_node(
                    dataset_type.name, self.empty_data_id, is_global_init_output=True
                )
                ref = self.empty_dimensions_datasets.outputs_in_the_way.get(dataset_key)
                if ref is None:
                    ref = DatasetRef(dataset_type, self.empty_data_id, run=self.output_run)
                full_skeleton.set_dataset_ref(ref, dataset_key)
            # Remove dataset nodes with no edges that are not global init
            # outputs, which are generally overall-inputs whose original quanta
            # end up skipped or with no work to do (we can't remove these along
            # with the quanta because no quantum knows if its the only
            # consumer).
            full_skeleton.remove_orphan_datasets()
            if attach_datastore_records:
                self._attach_datastore_records(full_skeleton)
        return full_skeleton

    @abstractmethod
    def process_subgraph(self, subgraph: PipelineGraph) -> QuantumGraphSkeleton:
        """Build the rough structure for an independent subset of the
        `.QuantumGraph` and query for relevant existing datasets.

        Parameters
        ----------
        subgraph : `.pipeline_graph.PipelineGraph`
            Subset of the pipeline graph that should be processed by this call.
            This is always resolved and topologically sorted.  It should not be
            modified.

        Returns
        -------
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Class representing an initial quantum graph. See
            `.quantum_graph_skeleton.QuantumGraphSkeleton` docs for details.
            After this is returned, the object may be modified in-place in
            unspecified ways.

        Notes
        -----
        The `.quantum_graph_skeleton.QuantumGraphSkeleton` should associate
        `lsst.daf.butler.DatasetRef` objects with nodes for existing datasets.
        In particular:

        - `.quantum_graph_skeleton.QuantumGraphSkeleton.set_dataset_ref` must
          be used to associate existing datasets with all overall-input dataset
          nodes in the skeleton by querying `input_collections`.  This includes
          all standard input nodes and any prerequisite nodes added by the
          method (prerequisite nodes may also be left out entirely, as the base
          class can add them later, albeit possibly less efficiently).
        - `.quantum_graph_skeleton.QuantumGraphSkeleton.set_output_for_skip`
          must be used to associate existing datasets with output dataset nodes
          by querying `skip_existing_in`.
        - `.quantum_graph_skeleton.QuantumGraphSkeleton.add_output_in_the_way`
          must be used to associated existing outputs with output dataset nodes
          by querying `output_run` if `output_run_exists` is `True`. Note that
          the presence of such datasets is not automatically an error, even if
          `clobber` is `False`, as these may be quanta that will be skipped.

        `lsst.daf.butler.DatasetRef` objects for existing datasets with empty
        data IDs in all of the above categories may be found in the
        `empty_dimensions_datasets` attribute, as these are queried for prior
        to this call by the base class, but associating them with graph nodes
        is still this method's responsibility.

        Dataset types should never be components and should always use the
        "common" storage class definition in `pipeline_graph.DatasetTypeNode`
        (which is the data repository definition when the dataset type is
        registered).
        """
        raise NotImplementedError()

    @final
    @timeMethod
    def _resolve_task_quanta(self, task_node: TaskNode, skeleton: QuantumGraphSkeleton) -> list[QuantumKey]:
        """Process the quanta for one task in a skeleton graph to skip those
        that have already completed and add missing prerequisite inputs.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.

        Returns
        -------
        skipped_quanta : `list` [ `.quantum_skeleton_graph.QuantumKey` ]
            Keys of quanta that were already skipped because their metadata
            already exists in a ``skip_existing_in`` collections.

        Notes
        -----
        This method modifies ``skeleton`` in-place in several ways:

        - It associates a `lsst.daf.butler.DatasetRef` with all output datasets
          and drops input dataset nodes that do not have a
          `lsst.daf.butler.DatasetRef` already.  This ensures producing and
          consuming tasks start from the same `lsst.daf.butler.DatasetRef`.
        - It removes quantum nodes that are to be skipped because their outputs
          already exist in `skip_existing_in`.  It also marks their outputs
          as no longer in the way.
        - It adds prerequisite dataset nodes and edges that connect them to the
          quanta that consume them.
        """
        # Extract the helper object for the prerequisite inputs of this task,
        # and tell it to prepare to construct skypix bounds and timespans for
        # each quantum (these will automatically do nothing if nothing needs
        # those bounds).
        task_prerequisite_info = self.prerequisite_info[task_node.label]
        task_prerequisite_info.update_bounds()
        # Loop over all quanta for this task, remembering the ones we've
        # gotten rid of.
        skipped_quanta = []
        for quantum_key in skeleton.get_quanta(task_node.label):
            if self._skip_quantum_if_metadata_exists(task_node, quantum_key, skeleton):
                skipped_quanta.append(quantum_key)
                continue
            quantum_data_id = skeleton[quantum_key]["data_id"]
            skypix_bounds_builder = task_prerequisite_info.bounds.make_skypix_bounds_builder(quantum_data_id)
            timespan_builder = task_prerequisite_info.bounds.make_timespan_builder(quantum_data_id)
            self._update_quantum_for_adjust(
                quantum_key,
                skeleton,
                task_prerequisite_info,
                skypix_bounds_builder,
                timespan_builder,
            )
        for skipped_quantum in skipped_quanta:
            skeleton.remove_quantum_node(skipped_quantum, remove_outputs=False)
        return skipped_quanta

    @final
    @timeMethod
    def _adjust_task_quanta(
        self, task_node: TaskNode, skeleton: QuantumGraphSkeleton, skipped_quanta: list[QuantumKey]
    ) -> None:
        """Process the quanta for one task in a skeleton graph by calling the
        ``adjust_all_quanta`` and ``adjustQuantum`` hooks.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.
        skipped_quanta : `list` [ `.quantum_skeleton_graph.QuantumKey` ]
            Keys of quanta that were already skipped because their metadata
            already exists in a ``skip_existing_in`` collections.

        Notes
        -----
        This method modifies ``skeleton`` in-place in several ways:

        - It adds "inputs", "outputs", and "init_inputs" attributes to the
          quantum nodes, holding the same `NamedValueMapping` objects needed to
          construct an actual `Quantum` instances.
        - It removes quantum nodes whose
          `~PipelineTaskConnections.adjustQuantum` calls raise `NoWorkFound` or
          predict no outputs;
        - It removes the nodes of output datasets that are "adjusted away".
        - It removes the edges of input datasets that are "adjusted away".

        The difference between how adjusted inputs and outputs are handled
        reflects the fact that many quanta can share the same input, but only
        one produces each output.  This can lead to the graph having
        superfluous isolated nodes after processing is complete, but these
        should only be removed after all the quanta from all tasks have been
        processed.
        """
        # Give the task a chance to adjust all quanta together.  This
        # operates directly on the skeleton (via a the 'adjuster', which
        # is just an interface adapter).
        adjuster = QuantaAdjuster(task_node.label, self._pipeline_graph, skeleton, self.butler)
        task_node.get_connections().adjust_all_quanta(adjuster)
        # Loop over all quanta again, remembering those we get rid of in other
        # ways.
        no_work_quanta = []
        for quantum_key in skeleton.get_quanta(task_node.label):
            adjusted_outputs = self._adapt_quantum_outputs(task_node, quantum_key, skeleton)
            adjusted_inputs = self._adapt_quantum_inputs(task_node, quantum_key, skeleton)
            # Give the task's Connections class an opportunity to remove
            # some inputs, or complain if they are unacceptable.  This will
            # raise if one of the check conditions is not met, which is the
            # intended behavior.
            helper = AdjustQuantumHelper(inputs=adjusted_inputs, outputs=adjusted_outputs)
            quantum_data_id = skeleton[quantum_key]["data_id"]
            try:
                helper.adjust_in_place(task_node.get_connections(), task_node.label, quantum_data_id)
            except NoWorkFound as err:
                # Do not generate this quantum; it would not produce any
                # outputs.  Remove it and all of the outputs it might have
                # produced from the skeleton.
                try:
                    _, connection_name, _ = err.args
                    details = f"not enough datasets for connection {connection_name}."
                except ValueError:
                    details = str(err)
                self.log.debug(
                    "No work found for quantum %s of task %s: %s",
                    quantum_key.data_id_values,
                    quantum_key.task_label,
                    details,
                )
                no_work_quanta.append(quantum_key)
                continue
            if helper.outputs_adjusted:
                if not any(adjusted_refs for adjusted_refs in helper.outputs.values()):
                    # No outputs also means we don't generate this quantum.
                    self.log.debug(
                        "No outputs predicted for quantum %s of task %s.",
                        quantum_key.data_id_values,
                        quantum_key.task_label,
                    )
                    no_work_quanta.append(quantum_key)
                    continue
                # Remove output nodes that were not retained by
                # adjustQuantum.
                skeleton.remove_dataset_nodes(
                    self._find_removed(skeleton.iter_outputs_of(quantum_key), helper.outputs)
                )
            if helper.inputs_adjusted:
                if not any(bool(adjusted_refs) for adjusted_refs in helper.inputs.values()):
                    raise QuantumGraphBuilderError(
                        f"adjustQuantum implementation for {task_node.label}@{quantum_key.data_id_values} "
                        "returned outputs but no inputs."
                    )
                # Remove input dataset edges that were not retained by
                # adjustQuantum.  We can't remove the input dataset nodes
                # because some other quantum might still want them.
                skeleton.remove_input_edges(
                    quantum_key, self._find_removed(skeleton.iter_inputs_of(quantum_key), helper.inputs)
                )
            # Save the adjusted inputs and outputs to the quantum node's
            # state so we don't have to regenerate those data structures
            # from the graph.
            skeleton[quantum_key]["inputs"] = helper.inputs
            skeleton[quantum_key]["outputs"] = helper.outputs
        for no_work_quantum in no_work_quanta:
            skeleton.remove_quantum_node(no_work_quantum, remove_outputs=True)
        remaining_quanta = skeleton.get_quanta(task_node.label)
        self._resolve_task_init(task_node, skeleton, bool(skipped_quanta))
        message_terms = []
        if no_work_quanta:
            message_terms.append(f"{len(no_work_quanta)} had no work to do")
        if skipped_quanta:
            message_terms.append(f"{len(skipped_quanta)} previously succeeded")
        if adjuster.n_removed:
            message_terms.append(f"{adjuster.n_removed} removed by adjust_all_quanta")
        message_parenthetical = f" ({', '.join(message_terms)})" if message_terms else ""
        if remaining_quanta:
            self.log.info(
                "Generated %s for task %s%s.",
                _quantum_or_quanta(len(remaining_quanta)),
                task_node.label,
                message_parenthetical,
            )
        else:
            self.log.info(
                "Dropping task %s because no quanta remain%s.", task_node.label, message_parenthetical
            )
            skeleton.remove_task(task_node.label)
        if len(no_work_quanta) > len(remaining_quanta):
            only_overall_inputs = self._get_task_inputs_if_overall_only(task_node)
            self.log.warning(
                "More than half of %s quanta had no work to do given available inputs.\n"
                "A query constraint on one of %s may yield a much faster build.",
                task_node.label,
                only_overall_inputs,
            )

    def _get_task_inputs_if_overall_only(self, task_node: TaskNode) -> list[str] | None:
        """If the given task consumes only overall-inputs, return their names.
        Otherwise return `None`.
        """
        result: list[str] = []
        for read_edge in task_node.inputs.values():
            if self._pipeline_graph.producer_of(read_edge.parent_dataset_type_name) is None:
                result.append(read_edge.parent_dataset_type_name)
            else:
                return None
        return result

    def _skip_quantum_if_metadata_exists(
        self, task_node: TaskNode, quantum_key: QuantumKey, skeleton: QuantumGraphSkeleton
    ) -> bool:
        """Identify and drop quanta that should be skipped because their
        metadata datasets already exist.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        quantum_key : `QuantumKey`
            Identifier for this quantum in the graph.
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.

        Returns
        -------
        skipped : `bool`
            `True` if the quantum is being skipped and has been removed from
            the graph, `False` otherwise.

        Notes
        -----
        If the metadata dataset for this quantum exists in the
        `skip_existing_in` collections, the quantum will be skipped. This
        causes the quantum node to be removed from the graph.  Dataset nodes
        that were previously the outputs of this quantum will be associated
        with `lsst.daf.butler.DatasetRef` objects that were found in
        ``skip_existing_in``, or will be removed if there is no such dataset
        there.  Any output dataset in `output_run` will be removed from the
        "output in the way" category.
        """
        metadata_dataset_key = DatasetKey(
            task_node.metadata_output.parent_dataset_type_name, quantum_key.data_id_values
        )
        if skeleton.get_output_for_skip(metadata_dataset_key):
            # This quantum's metadata is already present in the the
            # skip_existing_in collections; we'll skip it.  But the presence of
            # the metadata dataset doesn't guarantee that all of the other
            # outputs we predicted are present; we have to check.
            for output_dataset_key in list(skeleton.iter_outputs_of(quantum_key)):
                # If this dataset was "in the way" (i.e. already in the
                # output run), it isn't anymore.
                skeleton.discard_output_in_the_way(output_dataset_key)
                if (output_ref := skeleton.get_output_for_skip(output_dataset_key)) is not None:
                    # Populate the skeleton graph's node attributes
                    # with the existing DatasetRef, just like a
                    # predicted output of a non-skipped quantum.
                    skeleton.set_dataset_ref(output_ref, output_dataset_key)
                else:
                    # Remove this dataset from the skeleton graph,
                    # because the quantum that would have produced it
                    # is being skipped and it doesn't already exist.
                    skeleton.remove_dataset_nodes([output_dataset_key])
            # Removing the quantum node from the graph will happen outside this
            # function.
            return True
        return False

    @final
    def _update_quantum_for_adjust(
        self,
        quantum_key: QuantumKey,
        skeleton: QuantumGraphSkeleton,
        task_prerequisite_info: PrerequisiteInfo,
        skypix_bounds_builder: SkyPixBoundsBuilder,
        timespan_builder: TimespanBuilder,
    ) -> None:
        """Update the quantum node in the skeleton by finding remaining
        prerequisite inputs and dropping regular inputs that we now know will
        not be produced.

        Parameters
        ----------
        quantum_key : `QuantumKey`
            Identifier for this quantum in the graph.
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.
        task_prerequisite_info : `~prerequisite_helpers.PrerequisiteInfo`
            Information about the prerequisite inputs to this task.
        skypix_bounds_builder : `~prerequisite_helpers.SkyPixBoundsBuilder`
            An object that accumulates the appropriate spatial bounds for a
            quantum.
        timespan_builder : `~prerequisite_helpers.TimespanBuilder`
            An object that accumulates the appropriate timespan for a quantum.

        Notes
        -----
        This first looks for outputs already present in the `output_run` (i.e.
        "in the way" in the skeleton); if it finds something and `clobber` is
        `True`, it uses that ref (it's not ideal that both the original dataset
        and its replacement will have the same UUID, but we don't have space in
        the quantum graph for two UUIDs, and we need the datastore records of
        the original there).  If `clobber` is `False`, `RuntimeError` is
        raised.  If there is no output already present, a new one with a random
        UUID is generated.  In all cases the dataset node in the skeleton is
        associated with a `lsst.daf.butler.DatasetRef`.
        """
        dataset_key: DatasetKey | PrerequisiteDatasetKey
        for dataset_key in skeleton.iter_outputs_of(quantum_key):
            dataset_data_id = skeleton.get_data_id(dataset_key)
            dataset_type_node = self._pipeline_graph.dataset_types[dataset_key.parent_dataset_type_name]
            if (ref := skeleton.get_output_in_the_way(dataset_key)) is None:
                ref = DatasetRef(dataset_type_node.dataset_type, dataset_data_id, run=self.output_run)
            elif not self.clobber:
                # We intentionally raise here, before running adjustQuantum,
                # because it'd be weird if we left an old potential output of a
                # task sitting there in the output collection, just because the
                # task happened to not actually produce it.
                raise OutputExistsError(
                    f"Potential output dataset {ref} already exists in the output run "
                    f"{self.output_run}, but clobbering outputs was not expected to be necessary."
                )
            skypix_bounds_builder.handle_dataset(dataset_key.parent_dataset_type_name, dataset_data_id)
            timespan_builder.handle_dataset(dataset_key.parent_dataset_type_name, dataset_data_id)
            skeleton.set_dataset_ref(ref, dataset_key)
        quantum_data_id = skeleton.get_data_id(quantum_key)
        # Process inputs already present in the skeleton - this should include
        # all regular inputs (including intermediates) and may include some
        # prerequisites.
        for dataset_key in list(skeleton.iter_inputs_of(quantum_key)):
            if (ref := skeleton.get_dataset_ref(dataset_key)) is None:
                # If the dataset ref hasn't been set either as an existing
                # input or as an output of an already-processed upstream
                # quantum, it's not going to be produced; remove it.
                skeleton.remove_dataset_nodes([dataset_key])
                continue
            skypix_bounds_builder.handle_dataset(dataset_key.parent_dataset_type_name, ref.dataId)
            timespan_builder.handle_dataset(dataset_key.parent_dataset_type_name, ref.dataId)
        # Query for any prerequisites not handled by process_subgraph.  Note
        # that these were not already in the skeleton graph, so we add them
        # now.
        skypix_bounds = skypix_bounds_builder.finish()
        timespan = timespan_builder.finish()
        for finder in task_prerequisite_info.finders.values():
            dataset_keys = []
            for ref in finder.find(
                self.butler, self.input_collections, quantum_data_id, skypix_bounds, timespan
            ):
                dataset_key = skeleton.add_prerequisite_node(ref)
                dataset_keys.append(dataset_key)
            skeleton.add_input_edges(quantum_key, dataset_keys)

    @final
    def _adapt_quantum_outputs(
        self,
        task_node: TaskNode,
        quantum_key: QuantumKey,
        skeleton: QuantumGraphSkeleton,
    ) -> NamedKeyDict[DatasetType, list[DatasetRef]]:
        """Adapt outputs for a preliminary quantum and put them into the form
        used by `~lsst.daf.butler.Quantum` and
        `~PipelineTaskConnections.adjustQuantum`.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        quantum_key : `QuantumKey`
            Identifier for this quantum in the graph.
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.

        Returns
        -------
        outputs : `~lsst.daf.butler.NamedKeyDict` [ \
                `~lsst.daf.butler.DatasetType`, `list` [ \
                `~lsst.daf.butler.DatasetRef` ] ]
            All outputs to the task, using the storage class and components
            defined by the task's own connections.
        """
        outputs_by_type: dict[str, list[DatasetRef]] = {}
        dataset_key: DatasetKey
        for dataset_key in skeleton.iter_outputs_of(quantum_key):
            ref = skeleton.get_dataset_ref(dataset_key)
            assert ref is not None, "Should have been added (or the node removed) in a previous pass."
            outputs_by_type.setdefault(dataset_key.parent_dataset_type_name, []).append(ref)
        adapted_outputs: NamedKeyDict[DatasetType, list[DatasetRef]] = NamedKeyDict()
        for write_edge in task_node.iter_all_outputs():
            dataset_type_node = self._pipeline_graph.dataset_types[write_edge.parent_dataset_type_name]
            edge_dataset_type = write_edge.adapt_dataset_type(dataset_type_node.dataset_type)
            adapted_outputs[edge_dataset_type] = [
                write_edge.adapt_dataset_ref(ref)
                for ref in sorted(outputs_by_type.get(write_edge.parent_dataset_type_name, []))
            ]
        return adapted_outputs

    @final
    def _adapt_quantum_inputs(
        self,
        task_node: TaskNode,
        quantum_key: QuantumKey,
        skeleton: QuantumGraphSkeleton,
    ) -> NamedKeyDict[DatasetType, list[DatasetRef]]:
        """Adapt input datasets for a preliminary quantum into the form used by
        `~lsst.daf.butler.Quantum` and
        `~PipelineTaskConnections.adjustQuantum`.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        quantum_key : `QuantumKey`
            Identifier for this quantum in the graph.
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.

        Returns
        -------
        inputs : `~lsst.daf.butler.NamedKeyDict` [ \
                `~lsst.daf.butler.DatasetType`, `list` [ \
                `~lsst.daf.butler.DatasetRef` ] ]
            All regular and prerequisite inputs to the task, using the storage
            class and components defined by the task's own connections.

        Notes
        -----
        This method trims input dataset nodes that are not already associated
        with a `lsst.daf.butler.DatasetRef`, and queries for prerequisite input
        nodes that do not exist.
        """
        inputs_by_type: dict[str, set[DatasetRef]] = {}
        dataset_key: DatasetKey | PrerequisiteDatasetKey
        for dataset_key in list(skeleton.iter_inputs_of(quantum_key)):
            ref = skeleton.get_dataset_ref(dataset_key)
            assert ref is not None, "Should have been added (or the node removed) in a previous pass."
            inputs_by_type.setdefault(dataset_key.parent_dataset_type_name, set()).add(ref)
        adapted_inputs: NamedKeyDict[DatasetType, list[DatasetRef]] = NamedKeyDict()
        for read_edge in task_node.iter_all_inputs():
            dataset_type_node = self._pipeline_graph.dataset_types[read_edge.parent_dataset_type_name]
            edge_dataset_type = read_edge.adapt_dataset_type(dataset_type_node.dataset_type)
            if (current_dataset_type := adapted_inputs.keys().get(edge_dataset_type.name)) is None:
                adapted_inputs[edge_dataset_type] = [
                    read_edge.adapt_dataset_ref(ref)
                    for ref in sorted(inputs_by_type.get(read_edge.parent_dataset_type_name, frozenset()))
                ]
            elif current_dataset_type != edge_dataset_type:
                raise NotImplementedError(
                    f"Task {task_node.label!r} has {edge_dataset_type.name!r} as an input via "
                    "two different connections, with two different storage class overrides. "
                    "This is not yet supported due to limitations in the Quantum data structure."
                )
            # If neither the `if` nor the `elif` above match, it means
            # multiple input connections have exactly the same dataset
            # type, and hence nothing to do after the first one.
        return adapted_inputs

    @final
    def _resolve_task_init(
        self, task_node: TaskNode, skeleton: QuantumGraphSkeleton, has_skipped_quanta: bool
    ) -> None:
        """Add init-input and init-output dataset nodes and edges for a task to
        the skeleton.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Pipeline graph description of the task.
        skeleton : `QuantumGraphSkeleton`
            In-progress quantum graph data structure to update in-place.
        has_skipped_quanta : `bool`
            Whether any of this task's quanta were skipped because they had
            already succeeded.
        """
        quanta = skeleton.get_quanta(task_node.label)
        task_init_key = TaskInitKey(task_node.label)
        if quanta:
            adapted_inputs: NamedKeyDict[DatasetType, DatasetRef] = NamedKeyDict()
            # Process init-inputs.
            input_keys: list[DatasetKey] = []
            for read_edge in task_node.init.iter_all_inputs():
                dataset_key = skeleton.add_dataset_node(
                    read_edge.parent_dataset_type_name, self.empty_data_id
                )
                skeleton.add_input_edge(task_init_key, dataset_key)
                if (ref := skeleton.get_dataset_ref(dataset_key)) is None:
                    try:
                        ref = self.empty_dimensions_datasets.inputs[dataset_key]
                    except KeyError:
                        raise InitInputMissingError(
                            f"Overall init-input dataset {read_edge.parent_dataset_type_name!r} "
                            f"needed by task {task_node.label!r} not found in input collection(s) "
                            f"{self.input_collections}."
                        ) from None
                    skeleton.set_dataset_ref(ref, dataset_key)
                for quantum_key in skeleton.get_quanta(task_node.label):
                    skeleton.add_input_edge(quantum_key, dataset_key)
                input_keys.append(dataset_key)
                adapted_ref = read_edge.adapt_dataset_ref(ref)
                adapted_inputs[adapted_ref.datasetType] = adapted_ref
            # Save the quantum-adapted init inputs to each quantum, and add
            # skeleton edges connecting the init inputs to each quantum.
            for quantum_key in skeleton.get_quanta(task_node.label):
                skeleton[quantum_key]["init_inputs"] = adapted_inputs
            # Process init-outputs.
            adapted_outputs: NamedKeyDict[DatasetType, DatasetRef] = NamedKeyDict()
            for write_edge in task_node.init.iter_all_outputs():
                dataset_key = skeleton.add_dataset_node(
                    write_edge.parent_dataset_type_name, self.empty_data_id
                )
                if (ref := self.empty_dimensions_datasets.outputs_in_the_way.get(dataset_key)) is None:
                    ref = DatasetRef(
                        self._pipeline_graph.dataset_types[write_edge.parent_dataset_type_name].dataset_type,
                        self.empty_data_id,
                        run=self.output_run,
                    )
                skeleton.set_dataset_ref(ref, dataset_key)
                skeleton.add_output_edge(task_init_key, dataset_key)
                adapted_ref = write_edge.adapt_dataset_ref(ref)
                adapted_outputs[adapted_ref.datasetType] = adapted_ref
            skeleton[task_init_key]["inputs"] = adapted_inputs
            skeleton[task_init_key]["outputs"] = adapted_outputs
        elif has_skipped_quanta:
            # No quanta remain for this task, but at least one quantum was
            # skipped because its outputs were present in the skip_existing_in
            # collections.  This means all init outputs should be present in
            # the skip_existing_in collections, too, and we need to put those
            # refs in the graph.
            for write_edge in task_node.init.iter_all_outputs():
                dataset_key = skeleton.add_dataset_node(
                    write_edge.parent_dataset_type_name, self.empty_data_id
                )
                if (ref := self.empty_dimensions_datasets.outputs_for_skip.get(dataset_key)) is None:
                    raise InitInputMissingError(
                        f"Init-output dataset {write_edge.parent_dataset_type_name!r} of skipped task "
                        f"{task_node.label!r} not found in skip-existing-in collection(s) "
                        f"{self.skip_existing_in}."
                    ) from None
                skeleton.set_dataset_ref(ref, dataset_key)
                # If this dataset was "in the way" (i.e. already in the output
                # run), it isn't anymore.
                skeleton.discard_output_in_the_way(dataset_key)
        # No quanta remain in this task, but none were skipped; this means
        # they all got pruned because of NoWorkFound conditions.  This
        # dooms all downstream quanta to the same fate, so we don't bother
        # doing anything with the task's init-outputs, since nothing is
        # going to consume them.

    @final
    @timeMethod
    def _find_empty_dimension_datasets(self) -> EmptyDimensionsDatasets:
        """Query for all dataset types with no dimensions, updating
        `empty_dimensions_datasets` in-place.

        This includes but is not limited to init inputs and init outputs.
        """
        inputs: dict[DatasetKey | PrerequisiteDatasetKey, DatasetRef] = {}
        outputs_for_skip: dict[DatasetKey, DatasetRef] = {}
        outputs_in_the_way: dict[DatasetKey, DatasetRef] = {}
        _, dataset_type_nodes = self._pipeline_graph.group_by_dimensions().get(self.universe.empty, ({}, {}))
        dataset_types = [node.dataset_type for node in dataset_type_nodes.values()]
        dataset_types.extend(self._global_init_output_types.values())
        for dataset_type in dataset_types:
            key = DatasetKey(dataset_type.name, self.empty_data_id.required_values)
            if (
                self._pipeline_graph.producer_of(dataset_type.name) is None
                and dataset_type.name not in self._global_init_output_types
            ):
                # Dataset type is an overall input; we always need to try to
                # find these.
                try:
                    ref = self.butler.find_dataset(dataset_type.name, collections=self.input_collections)
                except MissingDatasetTypeError:
                    ref = None
                if ref is not None:
                    inputs[key] = ref
            elif self.skip_existing_in:
                # Dataset type is an intermediate or output; need to find these
                # if only they're from previously executed quanta that we might
                # skip...
                try:
                    ref = self.butler.find_dataset(dataset_type.name, collections=self.skip_existing_in)
                except MissingDatasetTypeError:
                    ref = None
                if ref is not None:
                    outputs_for_skip[key] = ref
                    if ref.run == self.output_run:
                        outputs_in_the_way[key] = ref
            if self.output_run_exists and not self.skip_existing_starts_with_output_run:
                # ...or if they're in the way and would need to be clobbered
                # (and we haven't already found them in the previous block).
                try:
                    ref = self.butler.find_dataset(dataset_type.name, collections=[self.output_run])
                except MissingDatasetTypeError:
                    ref = None
                if ref is not None:
                    outputs_in_the_way[key] = ref
        return EmptyDimensionsDatasets(
            inputs=inputs, outputs_for_skip=outputs_for_skip, outputs_in_the_way=outputs_in_the_way
        )

    @final
    @timeMethod
    def _attach_datastore_records(self, skeleton: QuantumGraphSkeleton) -> None:
        """Add datastore records for all overall inputs to a preliminary
        quantum graph.

        Parameters
        ----------
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph to update in place.

        Notes
        -----
        On return, all quantum nodes in the skeleton graph will have a
        "datastore_records" attribute that is a mapping from datastore name
        to `lsst.daf.butler.DatastoreRecordData`, as used by
        `lsst.daf.butler.Quantum`.
        """
        self.log.info("Fetching and attaching datastore records for all overall inputs.")
        overall_inputs = skeleton.extract_overall_inputs()
        exported_records = self.butler._datastore.export_records(overall_inputs.values())
        for task_label in self._pipeline_graph.tasks:
            if not skeleton.has_task(task_label):
                continue
            self.log.verbose("Fetching and attaching datastore records for task %s.", task_label)
            task_init_key = skeleton.get_task_init_node(task_label)
            init_input_ids = {
                ref.id
                for dataset_key in skeleton.iter_inputs_of(task_init_key)
                if (ref := overall_inputs.get(dataset_key)) is not None
            }
            init_records = {}
            if init_input_ids:
                for datastore_name, records in exported_records.items():
                    matching_records = records.subset(init_input_ids)
                    if matching_records is not None:
                        init_records[datastore_name] = matching_records
            skeleton[task_init_key]["datastore_records"] = init_records
            for quantum_key in skeleton.get_quanta(task_label):
                quantum_records = {}
                input_ids = {
                    ref.id
                    for dataset_key in skeleton.iter_inputs_of(quantum_key)
                    if (ref := overall_inputs.get(dataset_key)) is not None
                }
                if input_ids:
                    for datastore_name, records in exported_records.items():
                        matching_records = records.subset(input_ids)
                        if matching_records is not None:
                            quantum_records[datastore_name] = matching_records
                skeleton[quantum_key]["datastore_records"] = quantum_records

    @final
    @timeMethod
    def _construct_quantum_graph(
        self, skeleton: QuantumGraphSkeleton, metadata: Mapping[str, Any]
    ) -> QuantumGraph:
        """Construct a `.QuantumGraph` object from the contents of a
        fully-processed `.quantum_graph_skeleton.QuantumGraphSkeleton`.

        Parameters
        ----------
        skeleton : `.quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph.  Must have "init_inputs", "inputs", and
            "outputs" attributes on all quantum nodes, as added by
            `_resolve_task_quanta`, as well as a "datastore_records" attribute
            as added by `_attach_datastore_records`.
        metadata : `~collections.abc.Mapping`
            Flexible metadata to add to the graph.

        Returns
        -------
        quantum_graph : `.QuantumGraph`
            DAG describing processing to be performed.
        """
        from .graph import QuantumGraph

        self.log.info("Transforming graph skeleton into a QuantumGraph instance.")
        quanta: dict[TaskDef, set[Quantum]] = {}
        init_inputs: dict[TaskDef, Iterable[DatasetRef]] = {}
        init_outputs: dict[TaskDef, Iterable[DatasetRef]] = {}
        for task_def in self._pipeline_graph._iter_task_defs():
            if not skeleton.has_task(task_def.label):
                continue
            self.log.verbose("Transforming graph skeleton nodes for task %s.", task_def.label)
            task_node = self._pipeline_graph.tasks[task_def.label]
            task_init_key = skeleton.get_task_init_node(task_def.label)
            task_init_state = skeleton[task_init_key]
            init_datastore_records: dict[str, DatastoreRecordData] = task_init_state.get(
                "datastore_records", {}
            )
            init_inputs[task_def] = task_init_state["inputs"].values()
            init_outputs[task_def] = task_init_state["outputs"].values()
            quanta_for_task: set[Quantum] = set()
            for quantum_key in skeleton.get_quanta(task_node.label):
                quantum_state = skeleton[quantum_key]
                quantum_datastore_records: dict[str, DatastoreRecordData] = quantum_state.get(
                    "datastore_records", {}
                )
                quanta_for_task.add(
                    Quantum(
                        taskName=task_node.task_class_name,
                        taskClass=task_node.task_class,
                        dataId=quantum_state["data_id"],
                        initInputs=quantum_state["init_inputs"],
                        inputs=quantum_state["inputs"],
                        outputs=quantum_state["outputs"],
                        datastore_records=DatastoreRecordData.merge_mappings(
                            quantum_datastore_records, init_datastore_records
                        ),
                    )
                )
            quanta[task_def] = quanta_for_task

        registry_dataset_types: list[DatasetType] = [
            node.dataset_type for node in self._pipeline_graph.dataset_types.values()
        ]

        all_metadata = self.metadata.to_dict()
        all_metadata.update(metadata)
        global_init_outputs: list[DatasetRef] = []
        for dataset_key in skeleton.global_init_outputs:
            ref = skeleton.get_dataset_ref(dataset_key)
            assert ref is not None, "Global init input refs should be resolved already."
            global_init_outputs.append(ref)
        self.log.verbose("Invoking QuantumGraph class constructor.")
        result = QuantumGraph(
            quanta,
            metadata=all_metadata,
            universe=self.universe,
            initInputs=init_inputs,
            initOutputs=init_outputs,
            globalInitOutputs=global_init_outputs,
            registryDatasetTypes=registry_dataset_types,
        )
        self.log.info("Graph build complete.")
        return result

    @final
    @timeMethod
    def _construct_components(
        self,
        skeleton: QuantumGraphSkeleton,
        output: str | None,
        metadata: Mapping[str, Any] | None,
    ) -> PredictedQuantumGraphComponents:
        """Return quantum graph components from a completed skeleton.

        Parameters
        ----------
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Temporary data structure used by the builder to represent the
            graph.
        output : `str` or `None`, optional
            Output `~lsst.daf.butler.CollectionType.CHAINED` collection that
            combines the input and output collections.
        metadata : `~collections.abc.Mapping`, optional
            Mapping of JSON-friendly metadata.  Collection information, the
            current user, and the current timestamp are automatically
            included.

        Returns
        -------
        components : `.quantum_graph.PredictedQuantumGraphComponents`
            Components that can be used to construct a graph object and/or save
            it to disk.
        """
        from .quantum_graph import (
            PredictedDatasetModel,
            PredictedQuantumDatasetsModel,
            PredictedQuantumGraphComponents,
        )

        self.log.info("Transforming graph skeleton into PredictedQuantumGraph components.")
        components = PredictedQuantumGraphComponents(pipeline_graph=self._pipeline_graph)
        components.header.inputs = list(self.input_collections)
        components.header.output_run = self.output_run
        components.header.output = output
        if metadata is not None:
            components.header.metadata.update(metadata)
        components.dimension_data = DimensionDataAttacher(
            records=skeleton.get_dimension_data(),
            dimensions=self._pipeline_graph.get_all_dimensions(),
        )
        components.init_quanta.root = [
            PredictedQuantumDatasetsModel.model_construct(
                quantum_id=generate_uuidv7(),
                task_label="",
                outputs={
                    dataset_key.parent_dataset_type_name: [
                        PredictedDatasetModel.from_dataset_ref(
                            cast(DatasetRef, skeleton.get_dataset_ref(dataset_key))
                        )
                    ]
                    for dataset_key in skeleton.global_init_outputs
                },
            )
        ]
        for task_node in self._pipeline_graph.tasks.values():
            if not skeleton.has_task(task_node.label):
                continue
            self.log.verbose("Transforming graph skeleton nodes for task %s.", task_node.label)
            task_init_key = TaskInitKey(task_node.label)
            init_quantum_datasets = PredictedQuantumDatasetsModel.model_construct(
                quantum_id=generate_uuidv7(),
                task_label=task_node.label,
                inputs=self._make_predicted_datasets(
                    skeleton,
                    task_node.init.iter_all_inputs(),
                    skeleton.iter_inputs_of(task_init_key),
                ),
                outputs=self._make_predicted_datasets(
                    skeleton,
                    task_node.init.iter_all_outputs(),
                    skeleton.iter_outputs_of(task_init_key),
                ),
                datastore_records={
                    datastore_name: records.to_simple()
                    for datastore_name, records in skeleton[task_init_key]
                    .get("datastore_records", {})
                    .items()
                },
            )
            components.init_quanta.root.append(init_quantum_datasets)
            for quantum_key in skeleton.get_quanta(task_node.label):
                quantum_datasets = PredictedQuantumDatasetsModel.model_construct(
                    quantum_id=generate_uuidv7(),
                    task_label=task_node.label,
                    data_coordinate=list(skeleton.get_data_id(quantum_key).full_values),
                    inputs=self._make_predicted_datasets(
                        skeleton,
                        task_node.iter_all_inputs(),
                        skeleton.iter_inputs_of(quantum_key),
                    ),
                    outputs=self._make_predicted_datasets(
                        skeleton,
                        task_node.iter_all_outputs(),
                        skeleton.iter_outputs_of(quantum_key),
                    ),
                    datastore_records={
                        datastore_name: records.to_simple()
                        for datastore_name, records in skeleton[quantum_key]
                        .get("datastore_records", {})
                        .items()
                    },
                )
                components.quantum_datasets[quantum_datasets.quantum_id] = quantum_datasets
        self.log.verbose("Building the thin summary graph.")
        components.set_thin_graph()
        components.set_header_counts()
        self.log.info("Graph build complete.")
        return components

    @staticmethod
    def _make_predicted_datasets(
        skeleton: QuantumGraphSkeleton,
        edges: Iterable[Edge],
        dataset_keys: Iterable[DatasetKey | PrerequisiteDatasetKey],
    ) -> dict[str, list[PredictedDatasetModel]]:
        """Make the predicted quantum graph model objects that represent the
        datasets from an iterable of pipeline graph edges.

        Parameters
        ----------
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Temporary data structure used by the builder to represent the
            graph.
        edges : `~collections.abc.Iterable` [ `.pipeline_graph.Edge` ]
            Pipeline graph edges.
        dataset_keys : `~collections.abc.Iterable` [ \
                `.quantum_graph_skeleton.DatasetKey` or\
                `.quantum_graph_skeleton.PrerequisiteDatasetKey` ]
            All nodes in the skeleton that correspond to any of the given
            pipeline graph edges.

        Returns
        -------
        predicted_datasets : `dict` [ `str`, \
                `list` [ `.quantum_graph.PredictedDatasetModel` ] ]
            Mapping of dataset models, keyed by connection name.
        """
        from .quantum_graph import PredictedDatasetModel

        connection_names_by_dataset_type: defaultdict[str, list[str]] = defaultdict(list)
        result: dict[str, list[PredictedDatasetModel]] = {}
        for edge in edges:
            connection_names_by_dataset_type[edge.parent_dataset_type_name].append(edge.connection_name)
            result[edge.connection_name] = []

        for dataset_key in dataset_keys:
            connection_names = connection_names_by_dataset_type.get(dataset_key.parent_dataset_type_name)
            if connection_names is None:
                # Ignore if this isn't one of the connections we're processing
                # (probably an init-input), which would also be predecessor to
                # a quantum node, but should be handled separately.
                continue
            ref = skeleton.get_dataset_ref(dataset_key)
            assert ref is not None, "DatasetRefs should have already been added to skeleton."
            for connection_name in connection_names:
                result[connection_name].append(PredictedDatasetModel.from_dataset_ref(ref))
        for refs in result.values():
            refs.sort(key=operator.attrgetter("data_coordinate"))
        return result

    @staticmethod
    @final
    def _find_removed(
        original: Iterable[DatasetKey | PrerequisiteDatasetKey],
        adjusted: NamedKeyMapping[DatasetType, Sequence[DatasetRef]],
    ) -> set[DatasetKey | PrerequisiteDatasetKey]:
        """Identify skeleton-graph dataset nodes that have been removed by
        `~PipelineTaskConnections.adjustQuantum`.

        Parameters
        ----------
        original : `~collections.abc.Iterable` [ `DatasetKey` or \
                `PrerequisiteDatasetKey` ]
            Identifiers for the dataset nodes that were the original neighbors
            (inputs or outputs) of a quantum.
        adjusted : `~lsst.daf.butler.NamedKeyMapping` [ \
                `~lsst.daf.butler.DatasetType`, \
                `~collections.abc.Sequence` [ `lsst.daf.butler.DatasetType` ] ]
            Adjusted neighbors, in the form used by `lsst.daf.butler.Quantum`.

        Returns
        -------
        removed : `set` [ `DatasetKey` ]
            Datasets in ``original`` that have no counterpart in ``adjusted``.
        """
        result = set(original)
        for dataset_type, kept_refs in adjusted.items():
            parent_dataset_type_name, _ = DatasetType.splitDatasetTypeName(dataset_type.name)
            for kept_ref in kept_refs:
                # We don't know if this was a DatasetKey or a
                # PrerequisiteDatasetKey; just try both.
                result.discard(DatasetKey(parent_dataset_type_name, kept_ref.dataId.required_values))
                result.discard(PrerequisiteDatasetKey(parent_dataset_type_name, kept_ref.id.bytes))
        return result


@dataclasses.dataclass(eq=False, order=False)
class EmptyDimensionsDatasets:
    """Struct that holds the results of empty-dimensions dataset queries for
    `QuantumGraphBuilder`.
    """

    inputs: Mapping[DatasetKey | PrerequisiteDatasetKey, DatasetRef] = dataclasses.field(default_factory=dict)
    """Overall-input datasets found in `QuantumGraphBuilder.input_collections`.

    This may include prerequisite inputs.  It does include init-inputs.
    It does not include intermediates.
    """

    outputs_for_skip: Mapping[DatasetKey, DatasetRef] = dataclasses.field(default_factory=dict)
    """Output datasets found in `QuantumGraphBuilder.skip_existing_in`.

    It is unspecified whether this contains init-outputs; there is
    no concept of skipping at the init stage, so this is not expected to
    matter.
    """

    outputs_in_the_way: Mapping[DatasetKey, DatasetRef] = dataclasses.field(default_factory=dict)
    """Output datasets found in `QuantumGraphBuilder.output_run`.

    This includes regular outputs and init-outputs.
    """


def _quantum_or_quanta(n: int) -> str:
    """Correctly pluralize 'quantum' if needed."""
    return f"{n} quanta" if n != 1 else "1 quantum"



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("QuantumResourceUsage",)

import datetime

import numpy as np
import pydantic

from ._task_metadata import TaskMetadata


class QuantumResourceUsage(pydantic.BaseModel):
    """A struct holding the most frequently used resource usage metrics for
    executed quanta.
    """

    memory: float = pydantic.Field(json_schema_extra={"unit": "byte"})
    """Maximum memory usage of the process that ran this quantum, in bytes.

    This is derived from the maximum resident set size at the end of the
    quantum's execution, and is expected to only grow as multiple quanta are
    executed in the same process; that makes this a lower bound on the actual
    memory use of any particular quantum.
    """

    start: datetime.datetime
    """Start time (UTC), corresponding to the beginning of the "prep" period.
    """

    prep_time: float = pydantic.Field(json_schema_extra={"unit": "s"})
    """Wall-clock time in seconds spent preparing for the execution of this
    quantum.

    This includes checking for existing inputs, outputs, clobbering when
    necessary, and running the `..PipelineTaskConnections.adjustQuantum` hook.
    It does not include process startup times, import times, or butler
    initialization overheads.
    """

    init_time: float = pydantic.Field(json_schema_extra={"unit": "s"})
    """Wall-clock time in seconds spent initializing the task used to run this
    quantum.

    This includes the time spent reading init-inputs.
    """

    run_time: float = pydantic.Field(json_schema_extra={"unit": "s"})
    """Wall-clock time in seconds spent executing this quantum.

    This includes time spent reading inputs and writing outputs.  It does not
    include the time spent writing the special log and metadata datasets.
    """

    run_time_cpu: float = pydantic.Field(json_schema_extra={"unit": "s"})
    """CPU time in seconds spent executing this quantum.

    This includes time spent reading inputs and writing outputs, to the extent
    that those operations actually spend CPU time at all.  It does not include
    the time spent writing the special log and metadata datasets.
    """

    @property
    def total_time(self) -> float:
        """Total wall-clock time spent in "prep", "init", and "run"."""
        return self.prep_time + self.init_time + self.run_time

    @property
    def end(self) -> datetime.datetime:
        """End time (UTC), corresponding to the end of the "run" period."""
        return self.start + datetime.timedelta(seconds=self.total_time)

    @classmethod
    def from_task_metadata(cls, metadata: TaskMetadata) -> QuantumResourceUsage | None:
        """Extract resource usage information from task metadata.

        Parameters
        ----------
        metadata : `TaskMetadata`
            Metadata written by
            `.single_quantum_executor.SingleQuantumExecutor`.

        Returns
        -------
        resource_usage : `QuantumResourceUsage` or `None`
            Resource usage information for this quantum, or `None` if the
            expected fields were not found.
        """
        try:
            quantum_metadata = metadata["quantum"]
        except KeyError:
            return None
        end = datetime.datetime.fromisoformat(quantum_metadata["endUtc"])
        start = datetime.datetime.fromisoformat(quantum_metadata["prepUtc"])
        try:
            start_init = datetime.datetime.fromisoformat(quantum_metadata["initUtc"])
        except KeyError:
            start_init = end
        try:
            start_run = datetime.datetime.fromisoformat(quantum_metadata["startUtc"])
        except KeyError:
            start_run = end
        try:
            run_time_cpu = quantum_metadata["endCpuTime"] - quantum_metadata["startCpuTime"]
        except KeyError:
            run_time_cpu = 0.0
        return cls(
            memory=quantum_metadata["endMaxResidentSetSize"],
            start=start,
            prep_time=(start_init - start).total_seconds(),
            init_time=(start_run - start_init).total_seconds(),
            run_time=(end - start_run).total_seconds(),
            run_time_cpu=run_time_cpu,
        )

    @staticmethod
    def get_numpy_fields() -> dict[str, np.dtype]:
        """Return a mapping from field name to the `numpy.dtype` for that
        field.
        """
        return {
            "memory": np.dtype(np.float32),
            "start": np.dtype(np.float64),
            "prep_time": np.dtype(np.float32),
            "init_time": np.dtype(np.float32),
            "run_time": np.dtype(np.float32),
            "run_time_cpu": np.dtype(np.float32),
        }

    @staticmethod
    def get_units() -> dict[str, str | None]:
        """Return a mapping from field name to units.

        Units are astropy-compatible strings.
        """
        return {
            "memory": "byte",
            "start": "s",
            "prep_time": "s",
            "init_time": "s",
            "run_time": "s",
            "run_time_cpu": "s",
        }

    def get_numpy_row(self) -> tuple[object, ...]:
        """Convert this object to a `tuple` that can used to initialize a
        NumPy structured array with the fields from `get_numpy_fields`.
        """
        return (
            self.memory,
            self.start.timestamp(),
            self.prep_time,
            self.init_time,
            self.run_time,
            self.run_time_cpu,
        )



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("GetNodeText", "format_dimensions", "format_task_class", "get_node_symbol")

import itertools
import re
import textwrap
from collections.abc import Iterator

import networkx
import networkx.algorithms.community
from wcwidth import wcswidth  # type: ignore

from lsst.daf.butler import DimensionGroup

from .._nodes import NodeKey, NodeType
from ._merge import MergedNodeKey
from ._options import NodeAttributeOptions
from ._status_annotator import DatasetTypeStatusInfo, NodeStatusOptions, StatusColors, TaskStatusInfo

DisplayNodeKey = NodeKey | MergedNodeKey
"""Type alias for graph keys that may be original task, task init, or dataset
type keys, or a merge of several keys for display purposes.
"""


def strip_ansi(s: str) -> str:
    """Remove ANSI escape codes from a string, so that `wcswidth()` measures
    the real visible width of the string.

    Parameters
    ----------
    s : `str`
        String to strip of ANSI escape codes.

    Returns
    -------
    stripped : `str`
        String with ANSI escape codes removed.
    """
    # ANSI escape sequence remover
    ansi_escape = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", s)


def render_segment(f: float) -> str:
    """Convert a float into a string of blocks, rounding to the nearest whole
    number of columns.

    Parameters
    ----------
    f : `float`
        Number of columns to fill.

    Returns
    -------
    blocks : `str`
        String of blocks filling the specified number of columns.
    """
    return "█" * round(f)


def get_node_symbol(node: DisplayNodeKey, x: int | None = None) -> str:
    """Return a single-character symbol for a particular node type.

    Parameters
    ----------
    node : `DisplayNodeKey`
        Named tuple used as the node key.
    x : `str`, optional
        Ignored; may be passed for compatibility with the `Printer` class's
        ``get_symbol`` callback.

    Returns
    -------
    symbol : `str`
        A single-character symbol.
    """
    match node:
        case NodeKey(node_type=NodeType.TASK):
            return "■"
        case NodeKey(node_type=NodeType.DATASET_TYPE):
            return "○"
        case NodeKey(node_type=NodeType.TASK_INIT):
            return "▣"
        case MergedNodeKey(node_type=NodeType.TASK):
            return "▥"
        case MergedNodeKey(node_type=NodeType.DATASET_TYPE):
            return "◍"
        case MergedNodeKey(node_type=NodeType.TASK_INIT):
            return "▤"
    raise ValueError(f"Unexpected node key: {node} of type {type(node)}.")


class GetNodeText:
    """A callback for the `Printer` class's `get_text` callback that
    prints detailed information about a node and can defer long entries to
    a footnote.

    Parameters
    ----------
    xgraph : `networkx.DiGraph` or `networkx.MultiDiGraph`
        NetworkX export of a `.PipelineGraph` that is being displayed.
    options : `NodeAttributeOptions`
        Options for how much information to display.
    width : `int` or `None`
        Number of display columns that node text can occupy.  `None` for
        unlimited.
    """

    def __init__(
        self,
        xgraph: networkx.DiGraph | networkx.MultiDiGraph,
        options: NodeAttributeOptions,
        width: int | None,
    ):
        self.xgraph = xgraph
        self.options = options
        self.width = width
        self.deferred: list[tuple[str, tuple[str, str], list[str]]] = []

    def __call__(self, node: DisplayNodeKey, x: int, style: tuple[str, str]) -> str:
        """Return a line of text describing a node.

        Parameters
        ----------
        node : `DisplayNodeKey`
            Named tuple used as the node key.
        x : `int`
            Ignored; may be passed for compatibility with the `Printer` class's
            ``get_text`` callback.
        style : `tuple` [`str`, `str`]
            Tuple of ANSI color codes for overflow markers.
        """
        state = self.xgraph.nodes[node]
        has_status = "status" in state

        # Build description.
        description = self._build_description(node, state)

        # Possibly build progress bar to append to description.
        progress_portion = ""
        if has_status:
            progress_portion = self.format_node_status(description, state["status"])

        # Stitch together the final line, handling overflow if needed.
        final_line = self._stitch_node_text(description, progress_portion, style)
        return final_line

    def _build_description(self, node: DisplayNodeKey, state: dict) -> str:
        """Build the node description, possibly with additional details.

        Parameters
        ----------
        node : `DisplayNodeKey`
            Named tuple used as the node key.
        state : `dict`
            Node attributes.

        Returns
        -------
        description : `str`
            The node description.
        """
        terms = [f"{node}:" if self.options.has_details(node.node_type) else str(node)]
        # Optionally append dimension info.
        if self.options.dimensions and node.node_type != NodeType.TASK_INIT:
            terms.append(self.format_dimensions(state["dimensions"]))

        # Optionally append task class name.
        if self.options.task_classes and (
            node.node_type is NodeType.TASK or node.node_type is NodeType.TASK_INIT
        ):
            terms.append(self.format_task_class(state["task_class_name"]))

        # Optionally append storage class name.
        if self.options.storage_classes and node.node_type is NodeType.DATASET_TYPE:
            terms.append(state["storage_class_name"])

        description = " ".join(terms)

        return description

    def _stitch_node_text(self, description: str, progress_portion: str, style: tuple[str, str]) -> str:
        """Make the final line of node text to display given the description
        and possibly a progress portion, and handle overflow.

        It measures the total width of the description and progress portion,
        and if it exceeds the screen width, it truncates the description and
        appends a footnote.

        Parameters
        ----------
        description : `str`
            The node description.
        progress_portion : `str`
            The progress portion of the node.
        style : `tuple` [`str`, `str`]
            Tuple of ANSI color codes for overflow markers.

        Returns
        -------
        final_line : `str`
            The final line of text to display.
        """
        final_line = f"{description}{progress_portion}" if progress_portion else description
        total_len = wcswidth(strip_ansi(final_line))

        if self.width and total_len > self.width:
            overflow_index = f"[{len(self.deferred) + 1}]"
            overflow_marker = f"...{style[0]}{overflow_index}{style[1]}"

            avail_desc_width = (
                self.width - wcswidth(strip_ansi(progress_portion)) - wcswidth(strip_ansi(overflow_marker))
            )
            if avail_desc_width < 0:
                avail_desc_width = 0

            truncated_desc = description[:avail_desc_width] + overflow_marker
            self.deferred.append((overflow_index, style, [description]))

            return f"{truncated_desc}{progress_portion}"
        else:
            return final_line

    def format_node_status(self, description: str, status: TaskStatusInfo | DatasetTypeStatusInfo) -> str:
        """Format the status of a task node.

        Parameters
        ----------
        description : `str`
            The node description.
        status : `TaskStatusInfo` or `DatasetTypeStatusInfo`
            Holds status information for a task or dataset type.

        Returns
        -------
        formatted : `str`
            The formatted status string.
        """
        if not isinstance(self.options.status, NodeStatusOptions):
            raise ValueError(f"Invalid node status options: {self.options.status!r}.")

        return format_node_status(
            description,
            self.options.status,
            status,
            self.width,
        )

    def format_dimensions(self, dimensions: DimensionGroup) -> str:
        """Format the dimensions of a task or dataset type node.

        Parameters
        ----------
        dimensions : `~lsst.daf.butler.DimensionGroup`
            The dimensions to be formatted.

        Returns
        -------
        formatted : `str`
            The formatted dimension string.
        """
        return format_dimensions(self.options, dimensions)

    def format_task_class(self, task_class_name: str) -> str:
        """Format the type object for a task or task init node.

        Parameters
        ----------
        task_class_name : `str`
            The name of the task class.

        Returns
        -------
        formatted : `str`
            The formatted string.
        """
        return format_task_class(self.options, task_class_name)

    def format_deferrals(self, width: int | None) -> Iterator[str]:
        """Iterate over all descriptions that were truncated earlier and
        replace with footnote placeholders.

        Parameters
        ----------
        width : `int` or `None`
            Number of columns to wrap descriptions at.

        Returns
        -------
        deferrals : `collections.abc.Iterator` [ `str` ]
            Lines of deferred text, already wrapped.
        """
        indent = "  "
        for index, style, terms in self.deferred:
            yield f"{style[0]}{index}{style[1]}"
            for term in terms:
                if width:
                    yield from textwrap.wrap(term, width, initial_indent=indent, subsequent_indent=indent)
                else:
                    yield term


def format_dimensions(options: NodeAttributeOptions, dimensions: DimensionGroup) -> str:
    """Format the dimensions of a task or dataset type node.

    Parameters
    ----------
    options : `NodeAttributeOptions`
        Options for how much information to display.
    dimensions : `~lsst.daf.butler.DimensionGroup`
        The dimensions to be formatted.

    Returns
    -------
    formatted : `str`
        The formatted dimension string.
    """
    match options.dimensions:
        case "full":
            return str(dimensions.names)
        case "concise":
            redundant: set[str] = set()
            for a, b in itertools.permutations(dimensions.required, 2):
                if a in dimensions.universe[b].required:
                    redundant.add(a)
            kept = [d for d in dimensions.required if d not in redundant]
            assert dimensions.universe.conform(kept) == dimensions
            return f"{{{', '.join(kept)}}}"
        case False:
            return ""
    raise ValueError(f"Invalid display option for dimensions: {options.dimensions!r}.")


def format_task_class(options: NodeAttributeOptions, task_class_name: str) -> str:
    """Format the type object for a task or task init node.

    Parameters
    ----------
    options : `NodeAttributeOptions`
        Options for how much information to display.
    task_class_name : `str`
        The name of the task class.

    Returns
    -------
    formatted : `str`
        The formatted string.
    """
    match options.task_classes:
        case "full":
            return task_class_name
        case "concise":
            return task_class_name.split(".")[-1]
        case False:
            return ""
    raise ValueError(f"Invalid display option for task_classes: {options.task_classes!r}.")


def _build_progress_bar(
    description: str,
    prefix: str,
    suffix: str,
    segments: list[tuple[str, float]],
    width: int | None,
    colors: StatusColors,
    min_bar_width: int,
) -> str:
    """Shared helper that constructs a multi-segment progress bar.

    Parameters
    ----------
    description : `str`
        Main node description (used for measuring available space).
    prefix : `str`
        Text before the bar (e.g. ' · 42%▕').
    suffix : `str`
        Text after the bar (e.g. '▏exp: 107 ...').
    segments : `list` of `tuple` [`str`, `float`]
        Each tuple is (color_code, fractionOfBarWidth). We'll create these
        segments in sequence.
    width : `int` or None
        Overall maximum line width. None for unlimited. We'll still respect the
        minimum bar width.
    colors : `StatusColors`
        An instance containing .reset and color fields.
    min_bar_width : `int`
        Minimum number of display columns for the bar.

    Returns
    -------
    formatted : str
        The assembled prefix + bar + suffix, sized with respect to the width.
    """
    used_len = wcswidth(strip_ansi(prefix)) + wcswidth(strip_ansi(suffix))
    if width is not None:
        bar_width = max(width - wcswidth(strip_ansi(description)) - used_len, min_bar_width)
    else:
        bar_width = min_bar_width

    # Build bar from fractional segments.
    bar_str = ""
    total_cols = 0
    for ansi_color, fraction in segments:
        cols = round(bar_width * fraction)
        bar_str += f"{ansi_color}{render_segment(cols)}{colors.reset}"
        total_cols += cols

    # Pad the bar to the full width.
    if total_cols < bar_width:
        bar_str += " " * (bar_width - total_cols)

    return prefix + bar_str + suffix


def format_node_status(
    description: str,
    status_options: NodeStatusOptions,
    status: TaskStatusInfo | DatasetTypeStatusInfo,
    width: int | None,
) -> str:
    """Build a progress bar for a task or dataset type node.

    Parameters
    ----------
    description : `str`
        Node description for measuring leftover columns.
    status_options : `NodeStatusOptions`
        Options for node status visualization.
    status : `TaskStatusInfo` or `DatasetTypeStatusInfo`
        Holds status information for a task or dataset type.
    width : `int` or None
        Overall width limit (None => unlimited).

    Returns
    -------
    formatted : str
        The final prefix + bar + suffix line.
    """
    import dataclasses

    status_abbreviations = {
        "expected": "exp",
        "succeeded": "suc",
        "failed": "fail",
        "blocked": "blk",
        "ready": "rdy",
        "running": "run",
        "wonky": "wnk",
        "unknown": "unk",
        "produced": "prd",
    }

    colors = status_options.colors
    expected = status.expected
    done = status.succeeded if isinstance(status, TaskStatusInfo) else status.produced
    full_success = done == expected
    status_lookup = dataclasses.asdict(status)

    percent = 100.0 * done / expected if expected else 0.0
    total = float(expected) if expected else 1.0
    prefix = ""

    if status_options.display_percent or status_options.display_counts:
        if not status_options.visualize or (status_options.visualize and status_options.display_percent):
            full_success_color = colors.succeeded if isinstance(status, TaskStatusInfo) else colors.produced
            color_code = full_success_color if full_success else colors.failed
            prefix += f"{color_code} ▶ {colors.reset}"
        if status_options.display_percent:
            pct = round(percent)
            if pct == 100 and not full_success:
                pct == 99  # Avoid showing 100% if not fully successful.
            prefix += f"{pct}%"
        if not status_options.visualize and status_options.display_percent and status_options.display_counts:
            prefix += " | "

    if status_options.visualize:
        prefix += "▕"

    suffix_parts = []
    segments = []

    for key, value in status_lookup.items():
        if value is not None:
            color_code = getattr(colors, key)
            if status_options.display_counts:
                label = status_abbreviations[key] if status_options.abbreviate else key
                suffix_parts.append(f"{label}:{color_code}{value}{colors.reset}")
            if key != "expected":
                # Build a progress bar segment.
                segments.append((color_code, value / total))

    # Produce suffix from the parts.
    suffix = "▏" if status_options.visualize else ""
    suffix += " | ".join(suffix_parts)

    if status_options.visualize:
        return _build_progress_bar(
            description, prefix, suffix, segments, width, colors, status_options.min_bar_width
        )
    else:
        return f"{prefix}{suffix}"



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest

from lsst.pipe.base import QuantumSuccessCaveats


class QuantumSuccessCaveatsTestCase(unittest.TestCase):
    """Tests for the QuantumSuccessCaveats flag enum."""

    def test_from_adjust_quantum_no_work(self):
        """Test caveats for the case where adjustQuantum raises NoWorkFound."""
        caveats = QuantumSuccessCaveats.from_adjust_quantum_no_work()
        self.assertEqual(caveats.concise(), "*A")
        self.assertLessEqual(set(caveats.concise()), caveats.legend().keys())

    def test_concise(self):
        """Test the concise representation of caveats."""
        caveats = QuantumSuccessCaveats.ANY_OUTPUTS_MISSING | QuantumSuccessCaveats.UPSTREAM_FAILURE_NO_WORK
        self.assertEqual(caveats.concise(), "+U")
        self.assertLessEqual(set(caveats.concise()), caveats.legend().keys())
        caveats = (
            QuantumSuccessCaveats.ALL_OUTPUTS_MISSING
            | QuantumSuccessCaveats.ANY_OUTPUTS_MISSING
            | QuantumSuccessCaveats.UNPROCESSABLE_DATA
        )
        self.assertEqual(caveats.concise(), "*D")
        self.assertLessEqual(set(caveats.concise()), caveats.legend().keys())
        caveats = QuantumSuccessCaveats.ANY_OUTPUTS_MISSING | QuantumSuccessCaveats.PARTIAL_OUTPUTS_ERROR
        self.assertEqual(caveats.concise(), "+P")
        self.assertLessEqual(set(caveats.concise()), caveats.legend().keys())
        caveats = QuantumSuccessCaveats.NO_WORK
        self.assertEqual(caveats.concise(), "N")
        self.assertLessEqual(set(caveats.concise()), caveats.legend().keys())


if __name__ == "__main__":
    unittest.main()
